#
# Copyright (c) 2021 by Delphix. All rights reserved.
#
"""
Create and sync a dSource
"""
from os.path import basename

from delphixpy.v1_10_2 import exceptions
from dxi._lib import dlpx_exceptions as dxe
from dxi._lib import dx_logging as log
from dxi._lib import run_job
from dxi._lib import util
from dxi._lib.run_async import run_async
from dxi._lib.util import DXIConfigConstantsHelper
from dxi.database._internal_lib._ingest_lib import dsource_link_ase
from dxi.database._internal_lib._ingest_lib import dsource_link_mssql
from dxi.database._internal_lib._ingest_lib import dsource_link_oracle
from dxi.dxi_tool_base import DXIBase


class DXIProvisionDsourceConstants(object):
    """
    Class of common constants used by Provision dSource
    """

    SINGLE_THREAD = False
    POLL = 10
    CONFIG = DXIConfigConstantsHelper().get_config()
    LOG_FILE_PATH = DXIConfigConstantsHelper().get_logdir()
    ENGINE_ID = "default"
    PARALLEL = 5
    ACTION = None
    MODULE_NAME = __name__
    VDB_LIST_HEADER = []
    FORCE = False


class DXIDsource(DXIBase):
    def __init__(
        self,
        engine=None,
        log_file_path=DXIProvisionDsourceConstants.LOG_FILE_PATH,
        config=DXIProvisionDsourceConstants.CONFIG,
        poll=DXIProvisionDsourceConstants.POLL,
        single_thread=DXIProvisionDsourceConstants.SINGLE_THREAD,
        parallel=DXIProvisionDsourceConstants.PARALLEL,
        action=DXIProvisionDsourceConstants.ACTION,
        module_name=DXIProvisionDsourceConstants.MODULE_NAME,
    ):
        super().__init__(
            parallel=parallel,
            poll=poll,
            config=config,
            log_file_path=log_file_path,
            single_thread=single_thread,
            engine=engine,
            module_name=module_name,
        )
        self.dsource_name = ""
        self.db_passwd = ""
        self.db_user = ""
        self.dx_group = ""
        self.env_name = ""
        self.envinst = ""
        self.ip_addr = ""
        self.db_type = ""
        self.logsync_mode = ""
        self.port_num = ""
        self.num_connections = ""
        self.files_per_set = ""
        self.rman_channels = ""
        self.create_bckup = ""
        self.bck_file = ""
        self.db_install_path = ""
        self.backup_loc_passwd = ""
        self.backup_loc_user = ""
        self.load_from_backup = ""
        self.ase_passwd = ""
        self.ase_user = ""
        self.backup_path = ""
        self.sync_mode = ""
        self.source_user = ""
        self.stage_user = ""
        self.stage_repo = ""
        self.stage_instance = ""
        self.validated_sync_mode = ""
        self.logsync = False
        self.stage_env = ""
        self.initial_load_type = ""
        self.delphix_managed = ""
        self.source_env = ""
        self.action = action

    def ingest_mssql(
        self,
        dsource_name,
        db_passwd,
        db_user,
        dx_group,
        logsync,
        validated_sync_mode,
        initial_load_type,
        delphix_managed,
        stage_env,
        stage_instance,
        backup_path,
        backup_loc_passwd,
        backup_loc_user,
        source_env
    ):
        self.dsource_name = dsource_name
        self.db_passwd = db_passwd
        self.db_user = db_user
        self.dx_group = dx_group
        self.logsync = logsync
        self.validated_sync_mode = validated_sync_mode
        self.initial_load_type = initial_load_type
        self.delphix_managed = delphix_managed
        self.stage_env = stage_env
        self.stage_instance = stage_instance
        self.backup_path = backup_path
        self.backup_loc_passwd = backup_loc_passwd
        self.backup_loc_user = backup_loc_user
        self.source_env = source_env
        try:
            self._execute_operation(self.__ingest_mssql_helper)
            return True
        except Exception:
            return False

    @run_async
    def __ingest_mssql_helper(self, engine, dlpx_obj, single_thread):
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        if dlpx_obj.server_session:
            try:
                with dlpx_obj.job_mode(single_thread):
                    mssql_obj = dsource_link_mssql.DsourceLinkMssql(
                        dlpx_obj,
                        self.dsource_name,
                        self.db_passwd,
                        self.db_user,
                        self.dx_group,
                        self.logsync,
                        self.stage_env,
                        self.validated_sync_mode,
                        self.initial_load_type,
                        self.delphix_managed,
                    )

                    mssql_obj.link_mssql_dsource(
                        self.source_env,
                        self.stage_env,
                        self.stage_instance,
                        self.backup_path,
                        self.backup_loc_passwd,
                        self.backup_loc_user,
                        uuid="Any",
                    )
                    self._add_last_job_to_track(dlpx_obj)
                    run_job.track_running_jobs(engine, dlpx_obj, self.poll)
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                exceptions.JobError,
                exceptions.HttpError,
            ) as err:
                log.print_exception(
                    f"Error in {basename(__file__)}:"
                    f'{engine["hostname"]}\n ERROR: {err}'
                )
                raise err
        self._remove_session(dlpx_obj.server_session)

    def ingest_sybase(
        self,
        source_name,
        db_user,
        db_password,
        group,
        source_env,
        env_name,
        stage_repo,
        backup_path,
        backup_files,
        logsync,
        create_bckup,
    ):
        self.dsource_name = source_name
        self.db_passwd = db_password
        self.db_user = db_user
        self.dx_group = group
        self.logsync = logsync
        self.stage_repo = stage_repo
        self.backup_path = backup_path
        self.bck_file = backup_files
        self.create_bckup = create_bckup
        self.env_name = env_name
        self.source_env = source_env
        try:
            self._execute_operation(self.__ingest_sybase_helper)
            return True
        except Exception:
            return False

    @run_async
    def __ingest_sybase_helper(self, engine, dlpx_obj, single_thread):
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        if dlpx_obj.server_session:
            try:
                with dlpx_obj.job_mode(single_thread):
                    ase_obj = dsource_link_ase.DsourceLinkASE(
                        dlpx_obj,
                        self.dsource_name,
                        self.db_passwd,
                        self.db_user,
                        self.dx_group,
                        self.logsync,
                        self.stage_repo,

                    )
                    ase_obj.link_ase_dsource(
                        self.backup_path,
                        self.bck_file,
                        self.create_bckup,
                        self.env_name,
                        self.source_env
                    )
                    self._add_last_job_to_track(dlpx_obj)
                    run_job.track_running_jobs(engine, dlpx_obj, self.poll)
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                exceptions.JobError,
                exceptions.HttpError,
            ) as err:
                log.print_exception(
                    f"Error in {basename(__file__)}:"
                    f'{engine["hostname"]}\n ERROR: {err}'
                )
                raise err
        self._remove_session(dlpx_obj.server_session)

    def ingest_oracle(
        self,
        dsource_name,
        db_passwd,
        db_user,
        dx_group,
        logsync,
        env_name,
        db_install_path,
        ip_addr,
        port_num,
        rman_channels,
        files_per_set,
        num_connections,
    ):
        self.dsource_name = dsource_name
        self.db_passwd = db_passwd
        self.db_user = db_user
        self.dx_group = dx_group
        self.logsync = logsync
        self.env_name = env_name
        self.db_install_path = db_install_path
        self.ip_addr = ip_addr
        self.port_num = int(port_num)
        self.rman_channels = rman_channels
        self.files_per_set = files_per_set
        self.num_connections = num_connections
        try:
            self._execute_operation(self._ingest_oracle_helper)
            return True
        except Exception:
            return False

    @run_async
    def _ingest_oracle_helper(self, engine, dlpx_obj, single_thread):
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        if dlpx_obj.server_session:
            try:
                with dlpx_obj.job_mode(single_thread):
                    linked_ora = dsource_link_oracle.DsourceLinkOracle(
                        dlpx_obj=dlpx_obj,
                        dsource_name=self.dsource_name,
                        db_passwd=self.db_passwd,
                        db_user=self.db_user,
                        dx_group=self.dx_group,
                        logsync=self.logsync,
                        num_connections=self.num_connections,
                        files_per_set=self.files_per_set,
                        rman_channels=self.rman_channels,
                    )
                    linked_ora.get_or_create_ora_sourcecfg(
                        self.env_name,
                        self.db_install_path,
                        self.ip_addr,
                        port_num=int(self.port_num),
                    )
                    self._add_last_job_to_track(dlpx_obj)
                    run_job.track_running_jobs(engine, dlpx_obj, self.poll)
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                exceptions.JobError,
                exceptions.HttpError,
            ) as err:
                e = util.format_err_msg(
                    err,
                    operation="Dsource Ingestion",
                    engine_name=engine["hostname"],
                )
                log.print_exception(f"Error in {basename(__file__)}" f"{e}")
                # raise err
                self.failures = [True]
        self._remove_session(dlpx_obj.server_session)



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

"""
Provision Delphix Virtual Databases
"""
from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import database
from delphixpy.v1_10_2.web import environment
from delphixpy.v1_10_2.web import group
from delphixpy.v1_10_2.web import repository
from delphixpy.v1_10_2.web import sourceconfig
from delphixpy.v1_10_2.web import vo
from delphixpy.v1_10_2.web.environment.oracle import clusternode
from delphixpy.v1_10_2.web.source.operationTemplate import operationTemplate
from dxi._lib import dlpx_exceptions as dxe
from dxi._lib import dx_logging as log
from dxi._lib import dx_timeflow as dxtf
from dxi._lib import get_references as ref


def _add_hook(hooks_name, dlpx_obj):
    hook_references = []
    if hooks_name:
        for hook in hooks_name.split(","):
            template = ref.find_obj_by_name(
                dlpx_obj.server_session, operationTemplate, hook.strip()
            )
            hook_references.append(
                template.operation
            )  # Searching for operation object
    return hook_references


class _ProvisionVDBMixin(object):
    def _provision_ase_vdb(self, dlpx_obj):
        """
        Create a Sybase ASE VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        :param no_truncate_log: Don't truncate log on checkpoint
        :type no_truncate_log: bool
        :return:
        """
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        dxtf_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.ASEProvisionParameters()
        vdb_params.container = vo.ASEDBContainer()
        if self.no_truncate_log:
            vdb_params.truncate_log_on_checkpoint = True
        else:
            vdb_params.truncate_log_on_checkpoint = False
        vdb_params.container.group = group_ref
        vdb_params.container.name = self.name
        vdb_params.source = vo.ASEVirtualSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = True
        vdb_params.source_config = vo.ASESIConfig()
        vdb_params.source_config.database_name = self.name
        vdb_params.source_config.repository = ref.find_obj_by_name(
            dlpx_obj.server_session, repository, self.envinst
        ).reference
        vdb_params.timeflow_point_parameters = dxtf_obj.set_timeflow_point(
            source_obj,
            self.timestamp_type,
            self.timestamp,
            engine_session_ref=dlpx_obj.server_session,
        )
        vdb_params.timeflow_point_parameters.container = source_obj.reference
        log.print_info(f"{engine_name} provisioning {self.name}")
        database.provision(dlpx_obj.server_session, vdb_params)
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_mssql_vdb(self, dlpx_obj):
        """
        Create a MSSQL VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.MSSqlProvisionParameters()
        vdb_params.container = vo.MSSqlDatabaseContainer()
        vdb_params.container.group = group_ref
        vdb_params.container.name = self.name
        vdb_params.source = vo.MSSqlVirtualSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = False
        vdb_params.source_config = vo.MSSqlSIConfig()
        vdb_params.source_config.database_name = self.name

        vdb_params.source.parameters = {}
        vdb_params.source.operations = vo.VirtualSourceOperations()
        vdb_params = self._set_vdb_params(vdb_params, dlpx_obj)

        vdb_params.source_config.repository = ref.find_obj_by_name(
            dlpx_obj.server_session, repository, self.envinst
        ).reference
        vdb_params.timeflow_point_parameters = timeflow_obj.set_timeflow_point(  # noqa
            source_obj,
            self.timestamp_type,
            self.timestamp,
            engine_session_ref=dlpx_obj.server_session,
        )
        vdb_params.timeflow_point_parameters.container = source_obj.reference
        log.print_info(f"{engine_name} provisioning {self.name}")
        database.provision(dlpx_obj.server_session, vdb_params)
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_vfiles_vdb(self, dlpx_obj):
        """
        Create a vfiles VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vfiles_params = vo.AppDataProvisionParameters()
        vfiles_params.source = vo.AppDataVirtualSource()
        vfiles_params.source_config = vo.AppDataDirectSourceConfig()
        vfiles_params.source.allow_auto_vdb_restart_on_host_reboot = True
        vfiles_params.container = vo.AppDataContainer()
        vfiles_params.container.group = group_ref
        vfiles_params.container.name = self.name
        vfiles_params.source_config.name = self.name
        vfiles_params.source_config.path = f"{self.mntpoint}/{self.name}"
        vfiles_params.source_config.repository = ref.find_db_repo(
            dlpx_obj.server_session,
            "AppDataRepository",
            environment_obj.reference,
            "Unstructured Files",
        )
        vfiles_params.source.name = self.name
        vfiles_params.source.parameters = {}
        vfiles_params.source.operations = vo.VirtualSourceOperations()

        vfiles_params = self._set_vdb_params(vfiles_params)

        if self.timestamp_type is None:
            vfiles_params.timeflow_point_parameters = (
                vo.TimeflowPointSemantic()
            )
            vfiles_params.timeflow_point_parameters.container = (
                source_obj.reference
            )
            vfiles_params.timeflow_point_parameters.location = "LATEST_POINT"
        elif self.timestamp_type.upper() == "SNAPSHOT":
            try:
                dx_snap_params = timeflow_obj.set_timeflow_point(
                    source_obj,
                    self.timestamp_type,
                    self.timestamp,
                    engine_session_ref=dlpx_obj.server_session,
                )
            except exceptions.RequestError as err:
                raise dxe.DlpxException(
                    f"Could not set the timeflow point:\n{err}"
                )
            if dx_snap_params.type == "TimeflowPointSemantic":
                vfiles_params.timeflow_point_parameters = (
                    vo.TimeflowPointSemantic()
                )
                vfiles_params.timeflow_point_parameters.container = (
                    dx_snap_params.container
                )
                vfiles_params.timeflow_point_parameters.location = (
                    dx_snap_params.location
                )
            elif dx_snap_params.type == "TimeflowPointTimestamp":
                vfiles_params.timeflow_point_parameters = (
                    vo.TimeflowPointTimestamp()
                )
                vfiles_params.timeflow_point_parameters.timeflow = (
                    dx_snap_params.timeflow
                )
                vfiles_params.timeflow_point_parameters.timestamp = (
                    dx_snap_params.timestamp
                )
        log.print_info(f"{engine_name}: Provisioning {self.name}\n")
        try:
            database.provision(dlpx_obj.server_session, vfiles_params)
        except (exceptions.RequestError, exceptions.HttpError) as err:
            raise dxe.DlpxException(
                f"ERROR: Could not create vdb  {self.name}\n{err}"
            )
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_oracle_si_vdb(self, dlpx_obj):
        """
        Create an Oracle SI VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        try:
            vdb_obj = ref.find_obj_by_name(  # noqa
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{self.name} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.OracleProvisionParameters()
        vdb_params.open_resetlogs = True
        vdb_params.container = vo.OracleDatabaseContainer()
        vdb_params.container.group = group_ref
        vdb_params.container.name = self.name
        vdb_params.source = vo.OracleVirtualSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = False
        vdb_params.source.mount_base = self.mntpoint
        vdb_params.source_config = vo.OracleSIConfig()
        vdb_params.source_config.environment_user = (
            environment_obj.primary_user
        )
        vdb_params.source.operations = vo.VirtualSourceOperations()

        vdb_params = self._set_vdb_params(vdb_params, dlpx_obj)

        vdb_params.source_config.database_name = self.name
        vdb_params.source_config.unique_name = self.name
        vdb_params.source_config.instance = vo.OracleInstance()
        vdb_params.source_config.instance.instance_name = self.name
        vdb_params.source_config.instance.instance_number = 1
        vdb_params.source_config.repository = ref.find_db_repo(
            dlpx_obj.server_session,
            "OracleInstall",
            environment_obj.reference,
            self.envinst,
        )
        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        vdb_params.timeflow_point_parameters = timeflow_obj.set_timeflow_point(  # noqa
            source_obj,
            self.timestamp_type,
            self.timestamp,
            engine_session_ref=dlpx_obj.server_session,
        )
        log.print_info(f"{engine_name}: Provisioning {self.name}")
        try:
            database.provision(dlpx_obj.server_session, vdb_params)
        except (exceptions.RequestError, exceptions.HttpError) as err:
            raise dxe.DlpxException(
                f"ERROR: Could not create vdb  {self.name}\n{err}"
            )
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_oracle_mt_vdb(self, dlpx_obj):
        """
        Create an Oracle Multi Tenant VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        vdb_name = self.name
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        cdb_obj = ref.find_obj_by_name(
            dlpx_obj.server_session, sourceconfig, self.source_db
        )
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, vdb_name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.OracleMultitenantProvisionParameters()
        vdb_params.open_resetlogs = True
        vdb_params.container = vo.OracleDatabaseContainer()
        vdb_params.container.group = group_ref
        vdb_params.container.name = vdb_name
        vdb_params.source = vo.OracleVirtualPdbSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = True
        vdb_params.source.mount_base = self.mntpoint
        vdb_params.source_config = vo.OraclePDBConfig()
        vdb_params.source_config.database_name = vdb_name
        vdb_params.source_config.cdb_config = cdb_obj.cdb_config
        vdb_params.source.operations = vo.VirtualSourceOperations()

        vdb_params = self._set_vdb_params(vdb_params, dlpx_obj)

        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        vdb_params.timeflow_point_parameters = timeflow_obj.set_timeflow_point(  # noqa
            source_obj,
            self.timestamp_type,
            self.timestamp,
            engine_session_ref=dlpx_obj.server_session,
        )
        log.print_info(f"{engine_name}: Provisioning {vdb_name}")
        try:
            database.provision(dlpx_obj.server_session, vdb_params)
        except (exceptions.RequestError, exceptions.HttpError) as err:
            raise dxe.DlpxException(
                f"ERROR: Could not create vdb  {vdb_name}\n{err}"
            )
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_oracle_rac_vdb(self, dlpx_obj):
        """
        Create an Oracle rac VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        engine_name = list(dlpx_obj.dlpx_ddps)[0]
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.OracleProvisionParameters()
        vdb_params.open_resetlogs = True
        vdb_params.container = vo.OracleDatabaseContainer()
        vdb_params.container.group = group_ref
        vdb_params.container.name = self.name
        vdb_params.source = vo.OracleVirtualSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = False
        vdb_params.source.mount_base = self.mntpoint
        vdb_params.source_config = vo.OracleRACConfig()
        vdb_params.source_config.environment_user = (
            environment_obj.primary_user
        )
        vdb_params.source.operations = vo.VirtualSourceOperations()

        vdb_params = self._set_vdb_params(vdb_params, dlpx_obj.server_session)

        vdb_params.source_config.database_name = self.name
        vdb_params.source_config.unique_name = self.name
        instances = []
        number = 0
        for ora_node in self.nodes.split(":"):
            number += 1
            instances.append(
                {
                    "type": "OracleRACInstance",
                    "instanceName": f"{self.name}_{number}",
                    "instanceNumber": number,
                    "node": ref.find_obj_by_name(
                        dlpx_obj.server_session, clusternode, ora_node
                    ).reference,
                }
            )
        vdb_params.source_config.instances = instances
        vdb_params.source_config.repository = ref.find_db_repo(
            dlpx_obj.server_session,
            "OracleInstall",
            environment_obj.reference,
            self.envinst,
        )
        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        vdb_params.timeflow_point_parameters = timeflow_obj.set_timeflow_point(
            source_obj, self.timestamp_type, self.timestamp
        )
        log.print_info(f"{engine_name}: Provisioning {self.name}")
        try:
            database.provision(dlpx_obj.server_session, vdb_params)
        except (exceptions.RequestError, exceptions.HttpError) as err:
            raise dxe.DlpxException(
                f"ERROR: Could not create vdb  {self.name}\n{err}"
            )
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _provision_postgres_vdb(self, dlpx_obj):
        """
        Create a Postgres VDB
        :param dlpx_obj: DDP session object
        :type dlpx_obj: lib.GetSession.GetSession object
        """
        engine_name = list(dlpx_obj.dlpx_ddps)[0]

        group_ref, environment_obj, source_obj = self._find_group_source_environment_data(  # noqa
            dlpx_obj
        )
        try:
            vdb_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.name
            )
            raise dxe.DlpxObjectExists(f"{vdb_obj} exists.")
        except dxe.DlpxObjectNotFound:
            pass
        vdb_params = vo.AppDataProvisionParameters()
        vdb_params.container = vo.AppDataContainer()
        vdb_params.container.group = group_ref
        vdb_params.container.name = self.name
        vdb_params.source = vo.AppDataVirtualSource()
        vdb_params.source.allow_auto_vdb_restart_on_host_reboot = False
        vdb_params.source.mount_base = self.mntpoint
        vdb_params.source.parameters = {"postgresPort": int(self.port_num)}
        vdb_params.source_config = vo.AppDataDirectSourceConfig()
        vdb_params.source_config.environment_user = (
            environment_obj.primary_user
        )
        vdb_params.source_config.path = self.mntpoint
        vdb_params.source_config.name = self.name
        vdb_params.source_config.repository = ref.find_db_repo(
            dlpx_obj.server_session,
            "AppDataRepository",
            environment_obj.reference,
            self.envinst,
        )
        # "Unstructured Files",
        vdb_params.source.operations = vo.VirtualSourceOperations()

        vdb_params = self._set_vdb_params(vdb_params, dlpx_obj)

        timeflow_obj = dxtf.DxTimeflow(dlpx_obj.server_session)
        vdb_params.timeflow_point_parameters = timeflow_obj.set_timeflow_point(  # noqa
            source_obj,
            self.timestamp_type,
            self.timestamp,
            engine_session_ref=dlpx_obj.server_session,
        )
        log.print_info(f"{engine_name}: Provisioning {self.name}")
        try:
            database.provision(dlpx_obj.server_session, vdb_params)
        except (exceptions.RequestError, exceptions.HttpError) as err:
            raise dxe.DlpxException(
                f"ERROR: Could not create vdb  {self.name}\n{err}"
            )
        # Add the job into the jobs dictionary so we can track its progress
        self._add_last_job_to_track(dlpx_obj)

    def _find_group_source_environment_data(self, dlpx_obj):
        """
        :param dlpx_obj:
        :return:  group_ref (Group name where the VDB will be created),
        environment_obj (environment object where the VDB will be created),
        source_obj (Database object of the source)
        :rtype: tuple
        """
        group_ref = None
        environment_obj = None
        source_obj = None
        if self.group:
            group_ref = ref.find_obj_by_name(
                dlpx_obj.server_session, group, self.group
            ).reference
        if self.env_name:
            environment_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, environment, self.env_name
            )
        if self.source_db:
            source_obj = ref.find_obj_by_name(
                dlpx_obj.server_session, database, self.source_db
            )
        return group_ref, environment_obj, source_obj

    def _set_vdb_params(self, vdb_params, dlpx_obj):

        vdb_params.source.operations.pre_refresh = _add_hook(
            self.pre_refresh, dlpx_obj
        )
        vdb_params.source.operations.post_refresh = _add_hook(
            self.post_refresh, dlpx_obj
        )
        vdb_params.source.operations.pre_rollback = _add_hook(
            self.pre_rollback, dlpx_obj
        )
        vdb_params.source.operations.post_rollback = _add_hook(
            self.post_rollback, dlpx_obj
        )
        vdb_params.source.operations.configure_clone = _add_hook(
            self.configure_clone, dlpx_obj
        )

        return vdb_params



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#
from . import dlpx_exceptions
from . import dx_logging
from . import dx_timeflow
from . import get_references
from . import get_session
from . import click_util
from . import run_job
from . import dxi_constants



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

import click
from dxi._lib.util import boolean_based_system_exit
from dxi.config.dxi_config import DXIConfig
from dxi.config.dxi_config import DXIConfigConstants


@click.group(
    short_help="\b\nInitialize and Configure dxi.\n"
               "[init | encrypt]"
)
def config():
    """
    DXI Config related operations.
    """
    pass


# encrypt
@config.command()
@click.option(
    "--config-file",
    help="\b\nPath of the config file (including filename) \n"
    "that contains engine information",
    default=DXIConfigConstants.CONFIG,
)
@click.option(
    "--log-dir",
    help="\b Path to the log directory",
    default=DXIConfigConstants.LOG_DIR,
)
def encrypt(config_file, log_dir):
    """
    Encrypts username and password fields in config file.
    """
    temp_obj = DXIConfig(config=config_file, log_dir=log_dir)
    boolean_based_system_exit(temp_obj.encrypt())


# init
@config.command()
@click.option(
    "--config-file",
    help="\b\n Path to dxi config directory",
    default=DXIConfigConstants.CONFIG,
)
@click.option(
    "--log-dir",
    help="\b Path to the dxi log directory",
    default=DXIConfigConstants.LOG_DIR,
)
def init(config_file, log_dir):
    """
    Creates config and log directories.
    Adds sample dxtools.conf file into the config directory.
    """
    temp_obj = DXIConfig(config_file, log_dir)
    boolean_based_system_exit(temp_obj.init())


if __name__ == "__main__":
    encrypt()



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#
"""
Create and sync a dSource
"""

from os.path import basename

from delphixpy.v1_10_2 import exceptions
from dxi._lib import dlpx_exceptions as dxe
from dxi._lib import dx_logging as log
from dxi._lib import run_job
from dxi._lib.run_async import run_async
from dxi.database._internal_lib._ingest_lib import dsource_link_ase
from dxi.database._internal_lib._ingest_lib import dsource_link_mssql
from dxi.database._internal_lib._ingest_lib import dsource_link_oracle


class _ProvisionDsourceMixin(object):
    def _provision_dsource_helper(self, dlpx_obj):
        if self.db_type.lower() == "oracle":
            linked_ora = dsource_link_oracle.DsourceLinkOracle(
                dlpx_obj,
                self.dsource_name,
                self.db_passwd,
                self.db_user,
                self.dx_group,
                self.logsync,
                self.logsync_mode,
                self.db_type,
            )
            linked_ora.get_or_create_ora_sourcecfg(
                self.env_name,
                self.db_install_path,
                self.ip_addr,
                self.port_num,
            )
        elif self.db_type.lower() == "sybase":
            ase_obj = dsource_link_ase.DsourceLinkASE(
                dlpx_obj,
                self.dsource_name,
                self.db_passwd,
                self.db_user,
                self.dx_group,
                self.logsync,
                self.stage_repo,
                self.db_type,
            )
            ase_obj.link_ase_dsource(
                self.backup_path,
                self.bck_file,
                self.create_bckup,
                self.env_name,
            )
        elif self.db_type.lower() == "mssql":
            mssql_obj = dsource_link_mssql.DsourceLinkMssql(
                dlpx_obj,
                self.dsource_name,
                self.db_passwd,
                self.db_user,
                self.dx_group,
                self.db_type,
                self.logsync,
                self.validated_sync_mode,
                self.initial_load_type,
                self.delphix_managed,
            )
            mssql_obj.link_mssql_dsource(
                self.stage_env,
                self.stage_instance,
                self.backup_path,
                self.backup_loc_passwd,
                self.backup_loc_user,
                uuid="Any",
            )

    @run_async
    def _provision_dsource_runner(self, engine, dlpx_obj, single_thread):
        """
        This function is where we create our main workflow.
        Use the @run_async decorator to run this function asynchronously.
        The @run_async decorator to run multithreaded on Delphix Engines
        simultaneously
        :param engine: Dictionary of engines
        :type engine: dictionary
        :param single_thread: True - run single threaded,
                               False - run multi-thread
        :type single_thread: bool
        """
        err = None
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        if dlpx_obj.server_session:
            try:
                with dlpx_obj.job_mode(single_thread):
                    self._provision_dsource_helper(dlpx_obj)
                    run_job.track_running_jobs(engine, dlpx_obj, self.poll)
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                exceptions.JobError,
                exceptions.HttpError,
            ) as err:
                log.print_exception(
                    f"Error in {basename(__file__)}:"
                    f'{engine["hostname"]}\n ERROR: {err}'
                )
            raise err



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import database
from delphixpy.v1_10_2.web import vo
from dxi._lib import dlpx_exceptions
from dxi._lib import dx_logging
from dxi._lib import dx_timeflow
from dxi._lib import get_references
from dxi._lib import run_job
from dxi._lib import util
from dxi._lib.run_async import run_async


class _RefreshMixin(object):
    def _run_refresh_for_vdb(self, dlpx_obj):
        """
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        """
        for vdb_name in self.vdb.split(":"):
            dx_timeflow_obj = dx_timeflow.DxTimeflow(dlpx_obj.server_session)
            dx_logging.print_debug(f" Refreshing {vdb_name}")
            container_obj = get_references.find_obj_by_name(
                dlpx_obj.server_session, database, vdb_name
            )
            source_obj = get_references.find_source_by_db_name(
                dlpx_obj.server_session, vdb_name
            )
            # Sanity check to make sure our container object has a reference
            if container_obj.reference:
                try:
                    if (
                        source_obj.virtual is not True
                        or source_obj.staging is True
                    ):
                        dx_logging.print_exception(
                            f"{self.vdb} is not a virtual object.\n"
                        )
                    elif source_obj.runtime.enabled == "ENABLED":
                        dx_logging.print_debug(
                            f"INFO: Refreshing {self.vdb} "
                            f"to {self.time_stamp}\n"
                        )
                # This exception is raised if refreshing a vFiles VDB since
                # AppDataContainer does not have virtual,
                # staging or enabled attributes
                except AttributeError:
                    pass
            if source_obj.reference:
                try:
                    source_db = database.get(
                        dlpx_obj.server_session,
                        container_obj.provision_container,
                    )
                except (exceptions.RequestError, exceptions.JobError) as err:
                    raise dlpx_exceptions.DlpxException(
                        f"Encountered error while refreshing:"
                        f"{vdb_name}:\n{err}"
                    )

                if str(container_obj.reference).startswith("ORACLE"):
                    refresh_params = vo.OracleRefreshParameters()
                else:
                    refresh_params = vo.RefreshParameters()

                refresh_params.timeflow_point_parameters = dx_timeflow_obj.set_timeflow_point(  # noqa
                    source_db,
                    timestamp_type=self.time_stamp_type,
                    timestamp=self.time_stamp,
                    engine_session_ref=dlpx_obj.server_session,
                )
                try:
                    database.refresh(
                        dlpx_obj.server_session,
                        container_obj.reference,
                        refresh_params,
                    )
                    self._add_last_job_to_track(dlpx_obj)
                except (
                    dlpx_exceptions.DlpxException,
                    exceptions.RequestError,
                ) as err:
                    dx_logging.print_exception(
                        f"ERROR: Could not set timeflow point:{err}"
                    )
                    raise dlpx_exceptions.DlpxException(
                        f"ERROR: Could not set time flow point:{err}"
                    )
            # Don't do anything if the database is disabled
            else:
                dx_logging.print_debug(
                    f"INFO: {container_obj.name} is not enabled."
                    f" Refresh will not continue.\n"
                )

    def _refresh_validate_data(self):
        self.display_choices(self)
        dx_logging.print_debug("Validating inputs for refresh operation")
        if self.vdb and self.group:
            raise Exception("Name and Group both cannot be passed together.")
        if not (self.vdb or self.group):
            raise Exception("Either vdb name or group must be defined")

        if self.time_stamp_type.upper() not in ["TIME", "SNAPSHOT"]:
            raise Exception(
                "Possible values for Time stamp type: TIME|SNAPSHOT "
            )

    @run_async
    def _refresh_helper(self, engine, dlpx_obj, single_thread):
        """
        This function is where we create our main workflow.
        Use the @run_async decorator to run this function asynchronously.
        The @run_async decorator allows us to run against multiple
        Delphix Engine simultaneously
        :param engine: Dictionary of engines
        :type engine: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        self._refresh_validate_data()
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        try:
            with dlpx_obj.job_mode(single_thread):
                self._run_refresh_for_vdb(dlpx_obj)
                dx_logging.print_debug(f"All refreshes must be running now")
                run_job.track_running_jobs(
                    engine, dlpx_obj, poll=self.poll, failures=self.failures
                )
        except (Exception, BaseException) as err:
            e = util.format_err_msg(
                err,
                engine_name=engine["hostname"],
                operation="VDB/Vfile Refresh",
            )
            dx_logging.print_exception(f"Error in dxi_refresh\n" f"{e}")
            self.failures[0] = True



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

import click
from dxi._lib.util import boolean_based_system_exit
from dxi.branch.dxi_branch import DXIBranch
from dxi.branch.dxi_branch import DXIBranchConstants


@click.group(
    short_help="\b\nPerform Delphix Self-Service Branch Operations.\n"
               "[create | activate | delete | list] "
)
def branch():
    """
    Self-Service Branch operations
    """
    pass


@branch.command()
@click.option(
    "--branch-name", required=True, help="Name of the self-service branch."
)
@click.option(
    "--engine",
    default=DXIBranchConstants.ENGINE_ID,
    help="\b\nName of Delphix Engine in the config file.\n"
    "[Default: default]",
)
@click.option(
    "--single-thread",
    help="\b\nRun as a single thread\n"
    "[Default: {}]".format(DXIBranchConstants.SINGLE_THREAD),
    default=DXIBranchConstants.SINGLE_THREAD,
    is_flag=True,
)
@click.option(
    "--poll",
    type=click.INT,
    help="\b\nThe number of seconds to wait between job polls.\n"
    "[Default: {}s]".format(DXIBranchConstants.POLL),
    default=DXIBranchConstants.POLL,
)
@click.option(
    "--parallel",
    type=click.INT,
    help="Limit number of jobs to maxjob.",
    default=DXIBranchConstants.PARALLEL,
)
@click.option(
    "--config",
    help="The path to the dxtools.conf file.",
    default=DXIBranchConstants.CONFIG,
)
@click.option(
    "--log-file-path",
    help="The path to the logfile you want to use.",
    default=DXIBranchConstants.LOG_FILE_PATH,
)
def delete(
    engine, single_thread, parallel, poll, config, log_file_path, branch_name
):
    """
    Delete a self-service branch.
    """
    env_obj = DXIBranch(
        engine=engine,
        parallel=parallel,
        poll=poll,
        config_file=config,
        log_file_path=log_file_path,
        single_thread=single_thread,
    )
    boolean_based_system_exit(env_obj.delete(branch_name=branch_name))


@branch.command()
@click.option(
    "--branch-name", required=True, help="Name of the self-service branch."
)
@click.option(
    "--container-name",
    required=True,
    help="Name of the parent self-service container for the branch.",
)
@click.option(
    "--engine",
    default=DXIBranchConstants.ENGINE_ID,
    help="\b\nName of Delphix Engine in the config file.\n"
    "[Default: default]",
)
@click.option(
    "--single-thread",
    help="\b\nRun as a single thread\n"
    "[Default: {}]".format(DXIBranchConstants.SINGLE_THREAD),
    default=DXIBranchConstants.SINGLE_THREAD,
    is_flag=True,
)
@click.option(
    "--poll",
    type=click.INT,
    help="\b\nThe number of seconds to wait between job polls.\n"
    "[Default: {}s]".format(DXIBranchConstants.POLL),
    default=DXIBranchConstants.POLL,
)
@click.option(
    "--parallel",
    type=click.INT,
    help="Limit number of jobs to maxjob.",
    default=DXIBranchConstants.PARALLEL,
)
@click.option(
    "--config",
    help="The path to the dxtools.conf file.",
    default=DXIBranchConstants.CONFIG,
)
@click.option(
    "--log-file-path",
    help="The path to the logfile you want to use.",
    default=DXIBranchConstants.LOG_FILE_PATH,
)
def activate(
    engine,
    single_thread,
    parallel,
    poll,
    config,
    log_file_path,
    branch_name,
    container_name,
):
    """
    Activate a self-service branch.
    """
    env_obj = DXIBranch(
        engine=engine,
        parallel=parallel,
        poll=poll,
        config_file=config,
        log_file_path=log_file_path,
        single_thread=single_thread,
    )
    boolean_based_system_exit(
        env_obj.activate(
            branch_name=branch_name, container_name=container_name
        )
    )


@branch.command()
@click.option(
    "--engine",
    default=DXIBranchConstants.ENGINE_ID,
    help="\b\nName of Delphix Engine in the config file.\n"
    "[Default: default]",
)
@click.option(
    "--single-thread",
    help="\b\nRun as a single thread\n"
    "[Default: {}]".format(DXIBranchConstants.SINGLE_THREAD),
    default=DXIBranchConstants.SINGLE_THREAD,
    is_flag=True,
)
@click.option(
    "--poll",
    type=click.INT,
    help="\b\nThe number of seconds to wait between job polls.\n"
    "[Default: {}s]".format(DXIBranchConstants.POLL),
    default=DXIBranchConstants.POLL,
)
@click.option(
    "--parallel",
    type=click.INT,
    help="Limit number of jobs to maxjob.",
    default=DXIBranchConstants.PARALLEL,
)
@click.option(
    "--config",
    help="The path to the dxtools.conf file.",
    default=DXIBranchConstants.CONFIG,
)
@click.option(
    "--log-file-path",
    help="The path to the logfile you want to use.",
    default=DXIBranchConstants.LOG_FILE_PATH,
)
def list(engine, single_thread, parallel, poll, config, log_file_path):
    """
    List all self-service branches on an engine
    """
    env_obj = DXIBranch(
        engine=engine,
        parallel=parallel,
        poll=poll,
        config_file=config,
        log_file_path=log_file_path,
        single_thread=single_thread,
    )
    boolean_based_system_exit(env_obj.list())


@branch.command()
@click.option(
    "--branch-name",
    default=None,
    required=True,
    help="Name of the self-service branch.",
)
@click.option(
    "--container-name",
    default=None,
    required=True,
    help="Name of the parent self-service container.",
)
@click.option(
    "--template-name",
    default=None,
    help="Name of the parent self-service template.",
)
@click.option(
    "--bookmark-name",
    default=None,
    help="Name of the self-service bookmark to create the branch.",
)
@click.option(
    "--timestamp",
    default=None,
    help="Timestamp from the container's timeline to create the branch",
)
@click.option(
    "--engine",
    default=DXIBranchConstants.ENGINE_ID,
    help="\b\nName of Delphix Engine in the config file.\n"
    "[Default: default]",
)
@click.option(
    "--single-thread",
    help="\b\nRun as a single thread\n"
    "[Default: {}]".format(DXIBranchConstants.SINGLE_THREAD),
    default=DXIBranchConstants.SINGLE_THREAD,
    is_flag=True,
)
@click.option(
    "--poll",
    type=click.INT,
    help="\b\nThe number of seconds to wait between job polls.\n"
    "[Default: {}s]".format(DXIBranchConstants.POLL),
    default=DXIBranchConstants.POLL,
)
@click.option(
    "--parallel",
    type=click.INT,
    help="Limit number of jobs to maxjob.",
    default=DXIBranchConstants.PARALLEL,
)
@click.option(
    "--config",
    help="The path to the dxtools.conf file.",
    default=DXIBranchConstants.CONFIG,
)
@click.option(
    "--log-file-path",
    help="The path to the logfile you want to use.",
    default=DXIBranchConstants.LOG_FILE_PATH,
)
def create(
    branch_name,
    container_name,
    template_name,
    bookmark_name,
    timestamp,
    engine,
    single_thread,
    parallel,
    poll,
    config,
    log_file_path,
):
    """
    Create a new self-service branch.
    """
    env_obj = DXIBranch(
        engine=engine,
        parallel=parallel,
        poll=poll,
        config_file=config,
        log_file_path=log_file_path,
        single_thread=single_thread,
    )
    boolean_based_system_exit(
        env_obj.create(
            branch_name=branch_name,
            container_name=container_name,
            template_name=template_name,
            bookmark_name=bookmark_name,
            timestamp=timestamp,
        )
    )


if __name__ == "__main__":
    activate()



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#


from datetime import datetime
from pytz import UTC
from pytz import timezone
from dateutil import tz
from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import database
from delphixpy.v1_10_2.web import group
from delphixpy.v1_10_2.web import job
from delphixpy.v1_10_2.web import repository
from delphixpy.v1_10_2.web import source
from delphixpy.v1_10_2.web import sourceconfig
from delphixpy.v1_10_2.web import vo
from delphixpy.v1_10_2.web.service import time
from delphixpy.v1_10_2.web import selfservice
from delphixpy.v1_10_2.web import environment
from delphixpy.v1_10_2.web import snapshot

from . import dlpx_exceptions
from dxi._lib import dx_logging as log
from dxi._lib.util import convert_string_to_datetime


def convert_timestamp(engine, timestamp):
    """
    Convert timezone from Zulu/UTC to the Engine's timezone
    :param engine: A Delphix engine session object
    :type engine: lib.get_session.GetSession object
    :param timestamp: the timstamp in Zulu/UTC to be converted
    :type timestamp: str
    :return: Timestamp converted localtime
    """

    default_tz = tz.gettz("UTC")
    engine_tz = time.time.get(engine)
    try:
        convert_tz = tz.gettz(engine_tz.system_time_zone)
        utc = datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S")
        utc = utc.replace(tzinfo=default_tz)
        converted_tz = utc.astimezone(convert_tz)
        engine_local_tz = (
            f"{str(converted_tz.date())} "
            f"{str(converted_tz.time())} {str(converted_tz.tzname())}"
        )
        return engine_local_tz
    except TypeError:
        return None


def convert_timestamp_toiso(engine_session, timestamp):
    """
    Convert the provided timestamp to UTC.
    Args:
        engine_session: A Delphix engine session object
        timestamp: Timestamp in Engine's timezone.
    Returns:
        isotime: Iso Format timestamp in UTC Timezone.
    """
    utc_tz = tz.gettz("UTC")
    try:
        engine_time_config = time.time.get(engine_session)
        engine_tz_str = engine_time_config.system_time_zone
        engine_tz = timezone(engine_tz_str)
        ts = datetime.strptime(timestamp, '%Y-%m-%dT%H:%M:%S')
        ts_in_engine_tz = engine_tz.localize(ts)
        ts_in_utc = ts_in_engine_tz.astimezone(utc_tz)
        isotime = ts_in_utc.isoformat()
        return isotime
    except TypeError as err:
        log.print_exception(f"Error while converting the timstamp that was passed.")
        return None

def convert_timestamp_dbtz_to_utc(timestamp, db_tz, iso_format= False):
    """
    Convert timestamp from the DB timezone to UTC.
    Args:
        engine_session: A Delphix engine session object
        timestamp: Timestamp value provided by customer.
        db_tz: The timezone of the DB/Snapshot.
    Returns:
        converted_ts: timestamp in UTC timezone.
    """
    utc_tz = tz.gettz("UTC")
    try:
        if db_tz is None:
            raise Exception("Invalid timezone string.")
        database_tz = timezone(db_tz.split(",")[0])
        if not isinstance(timestamp, datetime):
            timestamp = convert_string_to_datetime(timestamp)
        if timestamp.tzinfo is None or timestamp.tzinfo.utcoffset(timestamp) is None: # noqa
            timestamp = database_tz.localize(timestamp)
        converted_ts = timestamp.astimezone(utc_tz)
        if iso_format:
            converted_ts = converted_ts.isoformat()
        return converted_ts
    except TypeError as err:
        log.print_exception(f"Error while converting the timstamp that was passed: {err}")
        return None
    except Exception as err:
        log.print_exception(f"Error while converting the timstamp that was passed: {err}")
        return None


def get_running_job(engine, object_ref):
    """
    Function to find a running job from the DB target reference.
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param object_ref: Reference to the object of the running job
    :type object_ref: str
    :return: reference of the running job(s)
    """
    try:
        return job.get_all(engine, target=object_ref, job_state="RUNNING")[
            0
        ].reference
    except IndexError:
        return None


def find_obj_by_name(engine, f_class, obj_name, env_name=None):
    """
    Function to find objects by name and object class
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow.
    :type f_class: Supported class type by Delphix
    :param obj_name: The name of the object
    :type obj_name: str
    :return: object of f_class type
    """
    if env_name:
        obj_list = f_class.get_all(engine, env_name)
    else:
        obj_list = f_class.get_all(engine)
    for obj in obj_list:
        if obj.name == obj_name:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(f"Object {obj_name} not found.")

def find_obj_by_name_v2(engine, f_class, obj_name, *args, **kwargs):
    """
    Function to find objects by name and object class
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow.
    :type f_class: Supported class type by Delphix
    :param obj_name: The name of the object
    :type obj_name: str
    :return: object of f_class type
    """
    obj_list = f_class.get_all(engine, *args, **kwargs)
    for obj in obj_list:
        if obj.name == obj_name:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(f"Object {obj_name} not found.")


def find_source_obj_by_containerref(engine, f_class, containerref):
    """
    Function to find objects by name and object class
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow.
    :type f_class: Supported class type by Delphix
    :param obj_name: The name of the object
    :type obj_name: str
    :return: object of f_class type
    """
    obj_list = f_class.get_all(engine)
    for obj in obj_list:
        if obj.container == containerref:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(f"Object with container reference {containerref} not found.")


def find_db_by_name_and_group(
        engine, group_name,dbname, exclude_js_container=False
):
    """
    Easy way to quickly find databases by group name and name
    :param engine: A Delphix DDP session object
    :type engine: lib.get_session.GetSession object
    :param group_name: Name of the group for the database
    :type group_name: str
    :param exclude_js_container: If set to true, search self-service
    containers
    :type exclude_js_container: bool
    :return: list of :py:class:`delphixpy.web.vo.Container`
    """
    # First search groups for the name specified and return its reference
    group_ref = find_obj_by_name(engine, group, group_name).reference
    if group_ref:
        databases = database.get_all(
            engine,
            group=group_ref,
            no_js_container_data_source=exclude_js_container,
        )
        for db in databases:
            if db.name == dbname:
                return db
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"Database: {dbname} was not found in group {group_name}"
    )


def find_obj_by_reference(engine, f_class, reference):
    """
    Function to find objects by reference and object class
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow.
    :type f_class: Supported class type by Delphix
    :param obj_name: The refere ce of the object
    :type reference: str
    :return: object of f_class type
    """
    obj_list = f_class.get_all(engine)
    for obj in obj_list:
        if obj.reference == reference:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"Object with reference {reference} not found."
    )

def find_source_by_db_name(engine, obj_name):
    """
    Function to find sources by database name and object class, and return
    object's reference as a string
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param obj_name: The name of the database object in Delphix
    :type obj_name: str
    :return: The parent DB object
    """
    for obj in database.get_all(engine):
        if obj.name == obj_name:
            source_obj = source.get_all(engine, database=obj.reference)
            return source_obj[0]
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"{obj_name} was not found on " f"engine {engine.address}.\n"
    )


def find_obj_name(engine, f_class, obj_reference):
    """
    Return the obj name from obj_reference

    :param engine: A Delphix DDP Session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow
    :type f_class: Supported class type by Delphix
    :param obj_reference: The object reference to retrieve the name
    :type obj_reference: str
    :return: str object name
    """
    try:
        obj_name = f_class.get(engine, obj_reference)
        return obj_name.name
    except (
        exceptions.RequestError,
        exceptions.JobError,
        exceptions.HttpError,
    ) as err:
        raise dlpx_exceptions.DlpxException(err)


def find_db_repo(engine, install_type, f_environment_ref, f_install_path):
    """
    Function to find database repository objects by environment reference and
    install path, and return the object's reference as a string
    You might use this function to find Oracle and PostGreSQL database repos.
    :param engine: A Delphix DDP session object
    :type engine: lib.GetSession.GetSession object
    :param install_type: Type of install - Oracle, or MSSQL
    :type install_type: str
    :param f_environment_ref: Reference of the environment for the repository
    :type f_install_path: str
    :param f_install_path: Path to the installation directory.
    :type f_install_path: str
    :return: delphixpy.web.vo.SourceRepository object
    """
    for obj in repository.get_all(engine, environment=f_environment_ref):
        if install_type == "OracleInstall":
            if (
                install_type == obj.type
                and obj.installation_home == f_install_path
            ):
                return obj.reference
        elif install_type == "MSSqlInstance":
            if (
                obj.type == install_type
                and obj.instance_name == f_install_path
            ):
                return obj.reference
        elif install_type == "AppDataRepository":
            try:
                if obj.type == install_type and obj.parameters['name'] == f_install_path:
                    return obj.reference
            except KeyError:
                if obj.type == install_type and obj.parameters['repositoryName'] == f_install_path:
                    return obj.reference
                else:
                    return None
        else:
            raise dlpx_exceptions.DlpxException(
                f"Only OracleInstall, AppDataRepository or MSSqlInstance "
                f"types are supported.\n"
            )


def find_sourceconfig(engine, sourceconfig_name, f_environment_ref):
    """
    Function to find database sourceconfig objects by environment reference,
    sourceconfig name (db name) and return the object
    You might use this function to find Oracle and PostGreSQL database
    sourceconfigs.
    :param engine: A Delphix DDP session object
    :type engine: lib.get_session.GetSession object
    :param sourceconfig_name: Name of source config, usually name of db
    instance (i.e. orcl)
    :type sourceconfig_name: str
    :param f_environment_ref: Reference of the environment for the repository
    :return: Type is determined by sourceonfig. Found in delphixpy.web.objects
    """
    for obj in sourceconfig.get_all(engine, environment=f_environment_ref):
        if obj.name == sourceconfig_name:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"No sourceconfig match found for type {sourceconfig_name}.\n"
    )


def find_all_databases_by_group(
    engine, group_name, exclude_js_container=False
):
    """
    Easy way to quickly find databases by group name
    :param engine: A Delphix DDP session object
    :type engine: lib.get_session.GetSession object
    :param group_name: Name of the group for the database
    :type group_name: str
    :param exclude_js_container: If set to true, search self-service
    containers
    :type exclude_js_container: bool
    :return: list of :py:class:`delphixpy.web.vo.Container`
    """
    # First search groups for the name specified and return its reference
    group_ref = find_obj_by_name(engine, group, group_name).reference
    if group_ref:
        databases = database.get_all(
            engine,
            group=group_ref,
            no_js_container_data_source=exclude_js_container,
        )
        return databases
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"No databases found in " f"group {group_name}.\n"
    )


def find_source_by_database(engine, database_obj):
    """
    The source tells us if the database is enabled/disabled, virtual,
    vdb/dSource, or is a staging database.
    :param engine: Delphix DDP Session object
    :type engine: lib.get_session.GetSession object
    :param database_obj: Delphix database object
    :type database_obj: delphixpy.web.vo.Container
    """
    source_obj = source.get_all(engine, database=database_obj.reference)
    # We'll just do a little sanity check here to ensure we only have a
    # 1:1 result.
    if not source_obj:
        raise dlpx_exceptions.DlpxObjectNotFound(
            f'{engine["hostname"]}: Did not find a source for '
            f"{database_obj.name}."
        )
    elif len(source_obj) > 1:
        raise dlpx_exceptions.DlpxException(
            f'{engine["hostname"]} More than one source returned for '
            f"{database_obj.name}"
        )
    return source_obj


def build_data_source_params(dlpx_obj, obj, data_source):
    """
    Builds the datasource parameters
    :param dlpx_obj: DDP session object
    :type dlpx_obj: lib.GetSession.GetSession object
    :param obj: object type to use when finding db
    :type obj: Type of object to build DS params
    :param data_source: Name of the database to use when building the
    parameters
    :type data_source: str
    """
    ds_params = vo.JSDataSourceCreateParameters()
    ds_params.source = vo.JSDataSource()
    ds_params.source.name = data_source
    try:
        db_obj = find_obj_by_name(dlpx_obj.server_session, obj, data_source)
        ds_params.container = db_obj.reference
        return ds_params
    except exceptions.RequestError as err:
        raise dlpx_exceptions.DlpxObjectNotFound(
            f"\nCould not find {data_source}\n{err}"
        )


def find_all_objects(engine, f_class):
    """
    Return all objects from a given class
    :param engine: A Delphix engine session object
    :type engine: lib.GetSession.GetSession object
    :param f_class: The objects class. I.E. database or timeflow.
    :return: list
    """
    try:
        return f_class.get_all(engine)
    except (exceptions.JobError, exceptions.HttpError) as err:
        raise dlpx_exceptions.DlpxException(
            f"{engine.address} Error encountered in {f_class}: {err}\n"
        )

def find_branch_by_name_and_container_ref(engine,branch_name, container_ref):
    """
    Finds a branch reference by name of the branch and
    referehcen of the container
    :param engine:
    :param container_ref:
    :return:
    """
    try:
        branches = find_all_objects(engine,selfservice.branch)
        if not branches:
            raise Exception
        for branch in branches:
            if branch.data_layout == container_ref and branch.name == branch_name: # noqa
                return branch
        raise Exception
    except Exception as err:
        raise dlpx_exceptions.DlpxObjectNotFound(f" Could not find branch:{branch_name}"
                                                 f" in container with ref: {container_ref}")



def find_obj_list(obj_lst, obj_name):
    """
    Function to find an object in a list of objects
    :param obj_lst: List containing objects from the get_all() method
    :type obj_lst: list
    :param obj_name: Name of the object to match
    :type obj_name: str
    :return: The named object, otherwise, DlpxObjectNotFound
    """
    for obj in obj_lst:
        if obj_name == obj.name:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(f"Did not find {obj_name}\n")

def find_obj_list_by_containername(obj_lst, container_name):
    """
    Function to find an object in a list of objects
    :param obj_lst: List containing objects from the get_all() method
    :type obj_lst: list
    :param obj_name: Name of the object to match
    :type obj_name: str
    :return: The named object, otherwise, DlpxObjectNotFound
    """
    for obj in obj_lst:
        if container_name == obj.container:
            return obj
    raise dlpx_exceptions.DlpxObjectNotFound(
        f"Did not find any object with containername:{container_name}\n")

def find_all_snapshots(engine_session, container_ref, timeflow_ref=None, from_date=None, to_date=None, timestamp=None):
    """
    Retrieves all snapshots for a specific dataset
    with the specified filters applied.
    Allowed filters:
        timeflow_ref: Retrieve snapshots from a specific timeflow.
        from_date: Retrive snapshots after this date.
        to_date: Retrieve snapshots before this date.
    Args:
        engine_session: Session to Delphix Engine
        container_ref: Reference to the dataset.
        timeflow_ref: Reference to timeflow
        from_date: Start date for snapshot.
        to_date: End date for snapshot.
    Returns:
        snapshot_obj: A TimeflowSnapshot object.
    Raises:
        Exception: Generic Exception
    """
    try:
        snapshots = snapshot.get_all(
            engine=engine_session,
            database=container_ref,
            from_date=from_date,
            to_date=to_date
        )
        return snapshots
    except Exception as err:
        raise Exception(f"There was an error retrieving snapshot for the specified timestamp.{err}")

def find_snapshot_by_timestamp(engine_session, container_ref, timestamp):
    """
    Returns the snapshots which can be used to provision
    to this timestamp for the given container.
    Allowed filters:
        timestamp: Timestamp to filter by.
        container_ref: VDB or vFile container object.
    Args:
        engine_session: Session to Delphix Engine
        container_ref: VDB or vFile container object.
        from_date: Start date for snapshot.
        to_date: End date for snapshot.
        timestamp: Timestamp to filter by.
    Returns:
        snapshot_obj: A TimeflowSnapshot object.
    Raises:
        Exception: Generic Exception
    """
    db_tz = None
    utc = UTC
    try:
        # unfortunately, we have to find the timezone of the DB
        # to do this, retrieve snapshots and find the timezone.
        all_snapshots = find_all_snapshots(engine_session,container_ref)
        if all_snapshots is None:
            raise Exception("No snapshots found for the VDB/vFile.")
        db_tz = all_snapshots[0].timezone
        ts_in_utc = convert_timestamp_dbtz_to_utc(timestamp, db_tz, iso_format=True)
        snapshots = snapshot.find_by_timestamp(
            engine=engine_session,
            container=container_ref,
            timestamp=ts_in_utc
        )
        #ts_in_utc = datetime.fromisoformat(ts_in_utc)
        #ts_in_utc= datetime.strptime(ts_in_utc,'%Y-%m-%dT%H:%M:%S.%fZ')
        return (db_tz, snapshots)
    except Exception as err:
        raise Exception(f"There was an error retrieving "
                        f"snapshot for the specified timestamp: {err}")


def find_db_timezone(engine_session, container_ref):
    """
    Returns the timezone for the given container.
    Args:
        engine_session: Session to Delphix Engine
        container_ref: VDB or vFile container object.
    Returns:
        timezone: Timezone string.
    Raises:
        Exception: Generic Exception
    """
    db_tz = None
    try:
        # TODO: Find a better way to do this?
        all_snapshots = find_all_snapshots(engine_session,container_ref)
        if all_snapshots is None:
            raise Exception("No snapshots found for the VDB/vFile.")
        for snapshot_obj in all_snapshots:
            db_tz = snapshot_obj.timezone
            break
        return db_tz
    except Exception as err:
        raise Exception(f"Unable to find the database timezone:{err}")



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#
from os.path import basename

from delphixpy.v1_10_2 import exceptions
from dxi._lib import dlpx_exceptions as dxe
from dxi._lib import dx_logging as log
from dxi._lib import run_job
from dxi._lib import util
from dxi._lib.run_async import run_async
from dxi._lib.util import DXIConfigConstantsHelper
from dxi.database._internal_lib._mixins._dxi_delete import _DeleteMixin
from dxi.database._internal_lib._mixins._dxi_provision_vdb import (
    _ProvisionVDBMixin,
)
from dxi.database._internal_lib._mixins._dxi_refresh import _RefreshMixin
from dxi.database._internal_lib._mixins._dxi_rewind import _RewindMixin
from dxi.dxi_tool_base import DXIBase


class DXIVdbConstants(object):
    """
    Define constants for DXIVDB class
    """

    SINGLE_THREAD = False
    POLL = 20
    CONFIG = DXIConfigConstantsHelper().get_config()
    LOG_FILE_PATH = DXIConfigConstantsHelper().get_logdir()
    ENGINE_ID = "default"
    TIME_STAMP_TYPE = "SNAPSHOT"
    TIME_STAMP = "LATEST"
    TIME_FLOW = None
    PARALLEL = 5
    FORCE = False
    TYPE = "vdb"
    DB_TYPE = None


class DXIVdb(
    DXIBase, _RefreshMixin, _DeleteMixin, _RewindMixin, _ProvisionVDBMixin
):
    """
    Implement operations for vdb datasets
    """

    def __init__(
        self,
        engine=DXIVdbConstants.ENGINE_ID,
        single_thread=DXIVdbConstants.SINGLE_THREAD,
        poll=DXIVdbConstants.POLL,
        config=DXIVdbConstants.CONFIG,
        log_file_path=DXIVdbConstants.LOG_FILE_PATH,
        parallel=DXIVdbConstants.PARALLEL,
        module_name=__name__,
    ):
        """
          :param engine: An Identifier of Delphix engine in dxtools.conf.
          :type engine: `str`
          :param single_thread: Run as a single thread.
                 False if running multiple threads.
          :type single_thread: `bool`
          :param poll: The number of seconds to wait between job polls
          :type poll: `int`
          :param config: The path to the dxtools.conf file
          :type config: `str`
          :param log_file_path: The path to the logfile you want to use.
          :type log_file_path: `str`
          :param all_dbs: Run against all database objects
          :type all_dbs: `bool`
        """
        super().__init__(
            poll=poll,
            config=config,
            log_file_path=log_file_path,
            single_thread=single_thread,
            engine=engine,
            module_name=module_name,
            parallel=parallel,
        )
        self.vdb = ""
        self.time_stamp_type = ""
        self.time_stamp = ""
        self.time_flow = ""
        self.name = ""
        self.force = ""
        self.type = ""
        self.source_db = ""
        self.group = ""
        self.db_type = ""
        self.env_name = ""
        self.mntpoint = ""
        self.timestamp = ""
        self.timestamp_type = ""
        self.pre_refresh = ""
        self.post_refresh = ""
        self.pre_rollback = ""
        self.post_rollback = ""
        self.configure_clone = ""
        self.envinst = ""
        self.action = ""
        self.no_truncate_log = ""
        self.port_num = ""

    def refresh(
        self,
        name,
        time_stamp_type=DXIVdbConstants.TIME_STAMP_TYPE,
        time_stamp=DXIVdbConstants.TIME_STAMP,
        time_flow=DXIVdbConstants.TIME_FLOW,
    ):
        """
        Refresh a Delphix VDB or vFile.
        :param name: VDBs name
        :type name: `str`
        :param time_stamp_type: Either SNAPSHOT or TIME
        :type time_stamp_type: `str`
        :param time_stamp: The Delphix semantic for the point in time on
            the source from which you want to refresh your VDB.
        :type time_stamp: `str`
        :param time_flow: Name of the timeflow to refresh a VDB
        """
        try:
            self.vdb = name
            self.time_stamp_type = time_stamp_type
            self.time_stamp = time_stamp
            self.time_flow = time_flow
            self._execute_operation(self._refresh_helper)
            return True
        except (Exception, BaseException):
            return False

    def delete(self, name, db_type="vdb", force=False):
        """
        Deletes a VDB or a list of VDBs from an engine

        :param name: Colon[:] separated names of the VDBs/dSources to delete.
        :type name: `str`
        :param db_type: Type of object being deleted. vdb | dsource
        :type db_type: `str`
        :param force: To force delete the objects
        :type force: `bool`
        """
        try:
            self.name = name
            self.force = force
            self.db_type = db_type
            self._execute_operation(self._delete_main_workflow)
            return True
        except (Exception, BaseException):
            return False

    def rewind(
        self,
        name,
        time_stamp_type=DXIVdbConstants.TIME_STAMP_TYPE,
        time_stamp=DXIVdbConstants.TIME_STAMP,
        database_type=None,
    ):
        """
        Rewind the given vdb

        :param name: VDBs name
        :type name: `str`
        :param time_stamp_type: Either SNAPSHOT or TIME
        :type time_stamp_type: `str`
        :param time_stamp: The Delphix semantic for the point in time on
                the source from which you want to refresh your VDB.
          :type time_stamp: `str`
          :param database_type: Type of database: oracle, mssql, ase, vfiles
          :type database_type: `str`
        """
        self.vdb = name
        self.time_stamp_type = time_stamp_type
        self.time_stamp = time_stamp
        self.db_type = database_type
        try:
            self._execute_operation(self._rewind_helper)
            return True
        except Exception:
            return False

    def provision_postgres(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_postgres_vdb
            )
            return True
        except Exception:
            return False

    def provision_vfiles(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_vfiles_vdb
            )
            return True
        except Exception:
            return False

    def provision_ase(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
        no_truncate_log=False,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        self.no_truncate_log = no_truncate_log
        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_ase_vdb
            )
            return True
        except Exception:
            return False

    def provision_mssql(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_mssql_vdb
            )
            return True
        except Exception:
            return False

    def provision_oracle_mt(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_oracle_mt_vdb
            )
            return True
        except Exception:
            return False

    def provision_oracle_si(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=5432,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num

        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_oracle_si_vdb
            )
            return True
        except Exception:
            return False

    def provision_oracle_rac(
        self,
        target_grp,
        source_db,
        db,
        env_name,
        nodes,
        prerefresh=False,
        postrefresh=False,
        prerollback=False,
        postrollback=False,
        configure_clone=False,
        port_num=1521,
        timestamp_type="SNAPSHOT",
        timestamp="LATEST",
        mntpoint="/mnt/provision",
        envinst=None,
    ):
        self.source_db = source_db
        self.name = db
        self.group = target_grp
        self.env_name = env_name
        self.mntpoint = mntpoint
        self.timestamp = timestamp
        self.timestamp_type = timestamp_type
        self.pre_refresh = prerefresh
        self.post_refresh = postrefresh
        self.pre_rollback = prerollback
        self.post_rollback = postrollback
        self.configure_clone = configure_clone
        self.envinst = envinst
        self.port_num = port_num
        self.nodes = nodes

        try:
            self._execute_operation(
                self._provision_vdb_helper, self._provision_oracle_rac_vdb
            )
            return True
        except Exception:
            return False

    @run_async
    def _provision_vdb_helper(
        self, engine, dlpx_obj, single_thread, function_ref=None
    ):
        """
        This function is where we create our main workflow.
        Use the @run_async decorator to run this function asynchronously.
        The @run_async decorator to run multithreaded on Delphix Engines
        simultaneously
        :param engine: Dictionary of engines
        :type engine: dictionary
        :param single_thread: True - run single threaded,
                               False - run multi-thread
        :type single_thread: bool
        """
        if not function_ref:
            self.failures = [True]
            raise Exception("Worker method is not defined for this operation")
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        if dlpx_obj.server_session:
            try:
                with dlpx_obj.job_mode(single_thread):
                    function_ref(dlpx_obj)
                    run_job.track_running_jobs(engine, dlpx_obj, self.poll)
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                exceptions.JobError,
                exceptions.HttpError,
            ) as err:
                e = util.format_err_msg(
                    err, operation="provision", engine_name=engine["hostname"]
                )
                log.print_exception(f"Error in {basename(__file__)}\n" f"{e}")
                self.failures = [True]
                # raise err
            except (dxe.DlpxObjectExists) as err:
                e = util.format_err_msg(
                    err.message,
                    operation="VDB/Vfile Provision",
                    engine_name=engine["hostname"],
                )
                log.print_exception(f"Error in {basename(__file__)}\n" f"{e}")
                self.failures = [True]
                # raise err



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

from delphixpy.v1_10_2.web import database
from delphixpy.v1_10_2.web import selfservice
from delphixpy.v1_10_2.web import user
from delphixpy.v1_10_2.web import vo
from dxi._lib import dx_logging
from dxi._lib import get_references
from dxi._lib import run_job
from dxi._lib.run_async import run_async
from dxi._lib.util import DXIConfigConstantsHelper
from dxi.dxi_tool_base import DXIBase
from tabulate import tabulate


class DXIContainerConstants(object):
    """
    Define constants for Self Service Container class and CLI usage
    """

    ALL_DBS = False
    SINGLE_THREAD = False
    POLL = 20
    CONFIG = DXIConfigConstantsHelper().get_config()
    LOG_FILE_PATH = DXIConfigConstantsHelper().get_logdir()
    BCK_FILE = None
    USE_BACKUP = False
    USE_BACKUP = False
    CREATE_BACKUP = False
    ENGINE_ID = "default"
    NAME = None
    GROUP = None
    PARALLEL = 5


class DXIContainer(DXIBase):
    """
    Create a snapshot a dSource or VDB
    """

    def __init__(
        self,
        engine=DXIContainerConstants.ENGINE_ID,
        single_thread=DXIContainerConstants.SINGLE_THREAD,
        config=DXIContainerConstants.CONFIG,
        log_file_path=DXIContainerConstants.LOG_FILE_PATH,
        poll=DXIContainerConstants.POLL,
    ):
        """
        :param engine: An Identifier of Delphix engine in dxtools.conf.
        :type engine: `str`
        :param single_thread: Run as a single thread.
            False if running multiple threads.
        :type single_thread: `bool`
        :param config: The path to the dxtools.conf file
        :type config: `str`
        :param log_file_path: The path to the logfile you want to use.
        :type log_file_path: `str`

        """
        super().__init__(
            config=config,
            log_file_path=log_file_path,
            single_thread=single_thread,
            engine=engine,
            poll=poll,
            module_name=__name__,
        )
        self.display_choices(self)
        self.container_name = ""
        self.template_name = ""
        self.database_name = ""
        self.keep_vdbs = False
        self.bookmark_name = ""
        self.owner_name = ""

    def list(self):
        """
        :return: Give all containers on a given engine
        :rtype: `list`
        """
        try:
            self._execute_operation(self.__list_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __list_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        :return: Give all containers on a given engine
        :rtype: `list`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        try:
            print_list = []
            with dlpx_obj.job_mode(single_thread):
                container_list = []
                header = [
                    "Name",
                    "Active Branch",
                    "Owner",
                    "Reference",
                    "Template",
                    "Last Updated",
                ]
                ss_containers = selfservice.container.get_all(
                    dlpx_obj.server_session
                )

                dx_logging.print_debug(header)

                for ss_container in ss_containers:
                    last_updated = get_references.convert_timestamp(
                        dlpx_obj.server_session, ss_container.last_updated[:-5]
                    )
                    dx_logging.print_debug(
                        f"{ss_container.name}, {ss_container.active_branch}, "
                        f"{ss_container.owner}, {ss_container.reference},"
                        f"{ss_container.template}, {last_updated}"
                    )
                    print_list.append(
                        [
                            ss_container.name,
                            ss_container.active_branch,
                            ss_container.owner,
                            ss_container.reference,
                            ss_container.template,
                            last_updated,
                        ]
                    )
                    container_list.append(
                        dict(
                            zip(
                                header,
                                [
                                    ss_container.name,
                                    ss_container.active_branch,
                                    ss_container.owner,
                                    ss_container.reference,
                                    ss_container.template,
                                    last_updated,
                                ],
                            )
                        )
                    )
                    self._add_last_job_to_track(dlpx_obj)
            run_job.track_running_jobs(
                engine_ref, dlpx_obj, poll=self.poll, failures=self.failures
            )
            print("\n")
            print(tabulate(print_list, headers=header, tablefmt="fancy_grid"))
            return container_list
        except Exception as err:
            dx_logging.print_exception(
                f"ERROR: SS Containers could not be listed. "
                f"The error was:\n{err}"
            )

    def create(self, container_name, template_name, db_name):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :param template_name:  Name of the JS Template to use for the container
        :type template_name: `str`
        :param db_name: Name of the child database(s) to use for the
                    SS Container
        :type db_name: `str`
        :return: created container reference
        :rtype: `str`
        """
        if not (container_name or template_name or db_name):
            dx_logging.print_exception("Some values are missing")
            return
        self.container_name = container_name
        self.template_name = template_name
        self.database_name = db_name
        try:
            self._execute_operation(self.__create_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __create_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        :return: created container reference
        :rtype: `str`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            try:
                ss_container_params = (
                    vo.JSDataContainerCreateWithoutRefreshParameters()
                )
                container_ds_lst = []
                for data_set in self.database_name.split(":"):
                    container_ds_lst.append(
                        get_references.build_data_source_params(
                            dlpx_obj, database, data_set
                        )
                    )
                ss_template_ref = get_references.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.template,
                    self.template_name,
                ).reference
                ss_container_params.template = ss_template_ref
                ss_container_params.timeline_point_parameters = (
                    vo.JSTimelinePointLatestTimeInput()
                )
                ss_container_params.timeline_point_parameters.sourceDataLayout = (  # noqa
                    ss_template_ref
                )
                ss_container_params.data_sources = container_ds_lst
                ss_container_params.name = self.container_name
                container_ref = selfservice.container.create(
                    dlpx_obj.server_session, ss_container_params
                )

                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
                return container_ref
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"Container {self.container_name} was not created. "
                    f"The error was:\n{str(err)}"
                )

    def delete(self, container_name, keep_vdbs=False):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :param keep_vdbs:  If set, deleting the container will not
            remove the underlying VDB
        :type keep_vdbs: `str`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        self.keep_vdbs = keep_vdbs
        try:
            self._execute_operation(self.__delete_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __delete_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            try:
                if self.keep_vdbs:
                    dx_logging.print_debug("VDB will not be deleted")
                    ss_container_params = vo.JSDataContainerDeleteParameters()
                    ss_container_params.delete_data_sources = False
                    selfservice.container.delete(
                        dlpx_obj.server_session,
                        get_references.find_obj_by_name(
                            dlpx_obj.server_session,
                            selfservice.container,
                            self.container_name,
                        ).reference,
                        ss_container_params,
                    )
                else:
                    selfservice.container.delete(
                        dlpx_obj.server_session,
                        get_references.find_obj_by_name(
                            dlpx_obj.server_session,
                            selfservice.container,
                            self.container_name,
                        ).reference,
                    )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"Container {self.container_name} was not deleted. "
                    f"The error was:\n{str(err)}"
                )

    def reset(self, container_name):
        """
        Undo the last refresh or restore operation

        :param container_name: Name of the SS Container
        :type container_name: `str`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        try:
            self._execute_operation(self.__reset_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __reset_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            try:
                selfservice.container.reset(
                    dlpx_obj.server_session,
                    get_references.find_obj_by_name(
                        dlpx_obj.server_session,
                        selfservice.container,
                        self.container_name,
                    ).reference,
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"ERROR: SS Container was not reset. "
                    f"The error was:\n{str(err)}"
                )

    def refresh(self, container_name):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        try:
            self._execute_operation(self.__refresh_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __refresh_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            try:
                selfservice.container.refresh(
                    dlpx_obj.server_session,
                    get_references.find_obj_by_name(
                        dlpx_obj.server_session,
                        selfservice.container,
                        self.container_name,
                    ).reference,
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"Container {self.container_name} was not refreshed. "
                    f"The error was:\n{str(err)}"
                )

    def restore(self, container_name, bookmark_name):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :param bookmark_name:  Name of the JS bookmark to
            restore for the container
        :type bookmark_name: `str`
        :return: True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        self.bookmark_name = bookmark_name
        try:
            self._execute_operation(self.__restore_container_helper)
            return True
        except Exception:
            return False

    @run_async
    def __restore_container_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            bookmark_params = vo.JSDataContainerRestoreParameters()
            bookmark_params.timeline_point_parameters = (
                vo.JSTimelinePointBookmarkInput()
            )
            bookmark_params.timeline_point_parameters.bookmark = get_references.find_obj_by_name(  # noqa
                dlpx_obj.server_session,
                selfservice.bookmark,
                self.bookmark_name,
            ).reference
            bookmark_params.force_option = False
            try:
                selfservice.container.restore(
                    dlpx_obj.server_session,
                    get_references.find_obj_by_name(
                        dlpx_obj.server_session,
                        selfservice.container,
                        self.container_name,
                    ).reference,
                    bookmark_params,
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"The container was not restored:\n{str(err)}"
                )

    def connection_info(self, container_name):
        """
        Lists hierarchy of a given container name
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :return: Hierarchy of a given container
        :rtype: `str`
        """
        self.container_name = container_name
        try:
            self._execute_operation(self.__connection_info_helper)
            return True
        except Exception:
            return False

    @run_async
    def __connection_info_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        headers = ["VDB", "Host", "DB Name", "DB Version", "JDBC"]
        connection_list = []
        with dlpx_obj.job_mode(single_thread):
            try:
                layout_ref = get_references.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.container,
                    self.container_name,
                ).reference
                for data_source in selfservice.datasource.get_all(
                    dlpx_obj.server_session, data_layout=layout_ref
                ):
                    data = []
                    db_name = get_references.find_obj_name(
                        dlpx_obj.server_session,
                        database,
                        data_source.container,
                    )

                    data.append(db_name)
                    data.append(data_source.runtime.host)
                    data.append(data_source.runtime.database_name)
                    data.append(data_source.runtime.version)

                    jdbc_str = ""
                    if hasattr(data_source.runtime, "instance_jdbc_string"):
                        jdbc_str = data_source.runtime.instance_jdbc_string
                    # For Oracle
                    elif hasattr(data_source.runtime, "jdbc_strings"):
                        jdbc_str = data_source.runtime.jdbc_strings

                    data.append(jdbc_str)
                    connection_list.append(data)

                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
                print(
                    tabulate(
                        connection_list, headers=headers, tablefmt="fancy_grid"
                    )
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"Failed to find connection info"
                    f"{self.container_name}:\n{str(err)}"
                )

    def add_owner(self, container_name, owner_name):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :param owner_name: Name of the JS Owner for the container
        :type owner_name: `str`
        :returns:  True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        self.owner_name = owner_name
        try:
            self._execute_operation(self.__add_owner_helper)
            return True
        except Exception:
            return False

    @run_async
    def __add_owner_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            owner_params = vo.JSDataContainerModifyOwnerParameters()
            try:
                owner_params.owner = get_references.find_obj_by_name(
                    dlpx_obj.server_session, user, self.owner_name
                ).reference
                selfservice.container.add_owner(
                    dlpx_obj.server_session,
                    get_references.find_obj_by_name(
                        dlpx_obj.server_session,
                        selfservice.container,
                        self.container_name,
                    ).reference,
                    owner_params,
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except (Exception, BaseException) as err:
                dx_logging.print_exception(
                    f"The user was not added from the container "
                    f"{self.container_name}:\n{str(err)}"
                )

    def remove_owner(self, container_name, owner_name):
        """
        :param container_name: Name of the SS Container
        :type container_name: `str`
        :param owner_name: Name of the JS Owner for the container
        :type owner_name: `str`
        :returns: True if execution succeeds, else False
        :rtype: `bool`
        """
        self.container_name = container_name
        self.owner_name = owner_name
        try:
            self._execute_operation(self.__remove_owner_helper)
            return True
        except Exception:
            return False

    @run_async
    def __remove_owner_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        with dlpx_obj.job_mode(single_thread):
            owner_params = vo.JSDataContainerModifyOwnerParameters()
            try:
                owner_params.owner = get_references.find_obj_by_name(
                    dlpx_obj.server_session, user, self.owner_name
                ).reference
                container_obj = get_references.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.container,
                    self.container_name,
                )
                selfservice.container.remove_owner(
                    dlpx_obj.server_session,
                    container_obj.reference,
                    owner_params,
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            except Exception as err:
                dx_logging.print_exception(
                    f"The user was not removed to container "
                    f"{self.container_name}. The error was:\n{err}"
                )



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

from os.path import basename

from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import database
from delphixpy.v1_10_2.web import source
from delphixpy.v1_10_2.web import vo
from dxi._lib import dlpx_exceptions
from dxi._lib import dx_logging
from dxi._lib import get_references
from dxi._lib import run_job
from dxi._lib.run_async import run_async
from dxi._lib.util import DXIConfigConstantsHelper
from dxi.dxi_tool_base import DXIBase


class SnapshotConstants(object):
    """
    Define constants for Snapshot class and CLI usage
    """

    ALL_DBS = False
    SINGLE_THREAD = False
    POLL = 20
    CONFIG = DXIConfigConstantsHelper().get_config()
    LOG_FILE_PATH = DXIConfigConstantsHelper().get_logdir()
    BCK_FILE = None
    USE_BACKUP = False
    CREATE_BACKUP = False
    ENGINE_ID = "default"
    NAME = None
    GROUP = None
    PARALLEL = 5


class DXISnapshot(DXIBase):
    """
    Create a snapshot a dSource or VDB
    """

    def __init__(
        self,
        db_name=SnapshotConstants.NAME,
        group=SnapshotConstants.GROUP,
        parallel=SnapshotConstants.PARALLEL,
        engine=SnapshotConstants.ENGINE_ID,
        poll=SnapshotConstants.POLL,
        config=SnapshotConstants.CONFIG,
        log_file_path=SnapshotConstants.LOG_FILE_PATH,
        all_dbs=SnapshotConstants.ALL_DBS,
        single_thread=SnapshotConstants.SINGLE_THREAD,
    ):
        """
        :param name: Name of object in Delphix to snapshot against
        :type name: `str`
        :param group: Name of group in Delphix to snapshot against.
        :type group: `str`
        :param parallel: Limit number of jobs to maxjob
        :type parallel: `int`
        :param engine: Alt Identifier of Delphix engine in dxtools.conf.
        :type engine: `str`
        :param poll: The number of seconds to wait between job polls
        :type poll: `int`
        :param config: The path to the dxtools.conf file
        :type config: `str`
        :param log_file_path: The path to the logfile you want to use.
        :type log_file_path: `str`
        :param all_dbs: Run against all database objects
        :type all_dbs: `bool`
        :param single_thread: Run as a single thread.
            False if running multiple threads.
        :type single_thread: `bool`
        """
        super().__init__(
            parallel=parallel,
            poll=poll,
            config=config,
            log_file_path=log_file_path,
            single_thread=single_thread,
            engine=engine,
            module_name=__name__,
        )
        self.all_dbs = all_dbs
        self.name = db_name
        self.group = group
        self.use_backup = SnapshotConstants.USE_BACKUP
        self.create_bckup = SnapshotConstants.CREATE_BACKUP
        self.bck_file = SnapshotConstants.BCK_FILE
        self.display_choices(self)
        self._validate_input()

    def _validate_input(self):
        if not (self.group or self.name or self.all_dbs):
            dx_logging.print_exception(
                "Invalid parameter is provided."
                " Select at least one option (GROUP|NAME|ALL_DBS)"
            )
            raise
        if (
            (self.group and self.name)
            or (self.name and self.all_dbs)
            or (self.name and self.all_dbs)
        ):
            dx_logging.print_exception(
                "Invalid parameter is provided. "
                "Please choose only single option in (GROUP|NAME|ALL_DBS)"
            )
            raise

    def create_snapshot(
        self, use_recent_backup=False, create_backup=False, backup_file=None
    ):
        """
        :param use_recent_backup: Snapshot using "Most Recent backup".
        :type use_recent_backup: `bool`
        :param create_backup:
            Create and ingest a new Sybase backup or copy-only MS SQL backup
        :type create_backup: `bool`
        :param backup_file: Name of the specific ASE Sybase backup file(s)
        :type backup_file: `str`
        :returns: Snapshot creation status
        :rtype: `bool`
        """
        self.use_backup = use_recent_backup
        self.create_bckup = create_backup
        self.bck_file = backup_file
        try:
            self._execute_operation(self._create_snapshot_helper)
            return True
        except Exception as err:
            dx_logging.print_exception(
                "Snapshot operation failed. {err}".format(err=str(err))
            )
            return False

    def _run_snapshot(self, dlpx_obj, datasets=None):
        """
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param datasets: Run against database objects
        :type datasets: `list`
        """
        sync_params = None

        for db_sync in datasets:
            try:
                db_source_info = get_references.find_obj_by_name(
                    dlpx_obj.server_session, source, db_sync
                )
                container_obj_ref = get_references.find_obj_by_name(
                    dlpx_obj.server_session, database, db_sync
                ).reference
            except dlpx_exceptions.DlpxObjectNotFound as err:
                raise err
            if db_source_info.staging:
                raise dlpx_exceptions.DlpxException(
                    f"{db_sync} is a staging " f"database. Cannot Sync.\n"
                )
            if db_source_info.runtime.enabled != "ENABLED":
                raise dlpx_exceptions.DlpxException(
                    f"{db_sync} is not enabled " f"database. Cannot Sync.\n"
                )
            if db_source_info.runtime.enabled == "ENABLED":
                # If the database is a dSource and a MSSQL type,
                # we need to tell Delphix how we want to sync the database.
                # Delphix will just ignore the extra parameters if it is a VDB,
                # so we will omit any extra code to check
                if db_source_info.type == "MSSqlLinkedSource":
                    if self.create_bckup:
                        sync_params = (
                            vo.MSSqlNewCopyOnlyFullBackupSyncParameters()
                        )
                        sync_params.compression_enabled = False
                    elif self.use_backup is True:
                        if self.bck_file:
                            sync_params = (
                                vo.MSSqlExistingSpecificBackupSyncParameters()
                            )
                            sync_params.backup_uuid = self.bck_file
                        else:
                            sync_params = (
                                vo.MSSqlExistingMostRecentBackupSyncParameters()  # noqa
                            )
                # Else if the database is a dSource and a ASE type,
                # we need also to tell Delphix how we want to sync the database
                # Delphix will just ignore the extra parameters if it is a VDB,
                # so we will omit any extra code to check
                elif db_source_info.type == "ASELinkedSource":
                    if self.use_backup is True:
                        sync_params = vo.ASELatestBackupSyncParameters()
                    elif self.bck_file:
                        sync_params = vo.ASESpecificBackupSyncParameters()
                        sync_params.backup_files = self.bck_file.split(" ")
                    elif self.create_bckup:
                        sync_params = vo.ASENewBackupSyncParameters()
                    else:
                        sync_params = vo.ASENewBackupSyncParameters()

                if sync_params:
                    database.sync(
                        dlpx_obj.server_session, container_obj_ref, sync_params
                    )
                else:
                    database.sync(dlpx_obj.server_session, container_obj_ref)
                # Add the job into the jobs dictionary so we can track
                # its progress
                self._add_last_job_to_track(dlpx_obj)

    def _get_db_list(self, dlpx_obj):
        """
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        """
        if self.name:
            self._run_snapshot(dlpx_obj, datasets=[self.name])
        else:
            databases = []
            if self.group:
                databases = get_references.find_all_databases_by_group(
                    dlpx_obj.server_session, self.group
                )
            elif self.all_dbs:
                # Grab all databases
                databases = database.get_all(
                    dlpx_obj.server_session, no_js_data_source=False
                )
            database_lst = [db_name.name for db_name in databases]

            self._run_snapshot(dlpx_obj, datasets=database_lst)

    @run_async
    def _create_snapshot_helper(self, engine, dlpx_obj, single_thread):
        """
        This function is where we create our main workflow.
        Use the @run_async decorator to run this function asynchronously.
        The @run_async decorator allows us to run against multiple
        Delphix Engine simultaneously
        :param engine: Dictionary of engines
        :type engine: `dict`
        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        :param single_thread: True - run single threaded, False -
            run multi-thread
        :type single_thread: `bool`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine)
        try:
            with dlpx_obj.job_mode(single_thread):
                self._get_db_list(dlpx_obj)
                run_job.track_running_jobs(
                    engine, dlpx_obj, poll=self.poll, failures=self.failures
                )
        except (
            dlpx_exceptions.DlpxObjectNotFound,
            exceptions.RequestError,
            exceptions.JobError,
            exceptions.HttpError,
        ) as err:
            dx_logging.print_exception(
                f'Error in {basename(__file__)}: {engine["hostname"]}\n{err}'
            )
            self.failures[0] = True



"""
Runs jobs passing a function as an argument. Thread safe.
"""
import time

from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import job
from dxi._lib import util

from . import dlpx_exceptions
from . import dx_logging

VERSION = "v.0.3.004"


def run_job(main_func, dx_obj, engine="default", single_thread=True):
    """
    This method runs the main_func asynchronously against all the
    delphix engines specified
    :param main_func: function to run against the DDP(s).
     In these examples, it's main_workflow().
    :type main_func: function
    :param dx_obj: Delphix session object from config
    :type dx_obj: lib.get_session.GetSession object
    :param engine: name of an engine, all or None
    :type engine: str
    :param single_thread: Run as single thread (True) or
                    multiple threads (False)
    :type single_thread: bool
    """
    threads = []
    # if engine ="all", run against every engine in config_file
    if engine == "all":
        dx_logging.print_info(f"Executing on all Delphix Engines")
        try:
            for delphix_ddp in dx_obj.dlpx_ddps:
                t = main_func(
                    dx_obj.dlpx_ddps[delphix_ddp], dx_obj, single_thread
                )
                threads.append(t)
                # TODO: Revisit threading logic
                # This sleep has been tactically added to prevent errors in the parallel
                # processing of operations across multiple engines
                time.sleep(1)
        except dlpx_exceptions.DlpxException as err:
            dx_logging.print_exception(f"Error encountered in run_job():{err}")
    elif engine == "default":
        try:
            for delphix_ddp in dx_obj.dlpx_ddps.keys():
                if dx_obj.dlpx_ddps[delphix_ddp]["default"] == "True":
                    dx_obj_default = dx_obj
                    dx_obj_default.dlpx_ddps = {
                        delphix_ddp: dx_obj.dlpx_ddps[delphix_ddp]
                    }
                    dx_logging.print_info(
                        "Executing on the default Delphix Engine"
                    )
                    t = main_func(
                        dx_obj.dlpx_ddps[delphix_ddp], dx_obj, single_thread
                    )
                    threads.append(t)
                break
        except TypeError as err:
            raise dlpx_exceptions.DlpxException(f"Error in run_job: {err}")
    else:
        # Test to see if the engine exists in config_file
        try:
            engine_ref = dx_obj.dlpx_ddps[engine]
            t = main_func(engine_ref, dx_obj, single_thread)
            threads.append(t)
            dx_logging.print_info(
                f"Executing on Delphix Engine: " f'{engine_ref["hostname"]}'
            )
        except (exceptions.RequestError, KeyError):
            raise dlpx_exceptions.DlpxException(
                f"\nERROR: Delphix Engine {engine} cannot be found. Please "
                f"check your input and try again."
            )
    if engine is None:
        raise dlpx_exceptions.DlpxException(
            f"ERROR: No default Delphix Engine found."
        )
    return threads


def _get_thread(main_func, engine, dx_obj, single_thread, helper_ref):
    if helper_ref:
        t = main_func(engine, dx_obj, single_thread, helper_ref)
    else:
        t = main_func(engine, dx_obj, single_thread)
    return t


def run_job_mt(
    main_func, dx_obj, engine="default", single_thread=True, helper_ref=None
):
    """
    This method runs the main_func asynchronously against all the
    delphix engines specified
    :param main_func: function to run against the DDP(s).
     In these examples, it's main_workflow().
    :type main_func: function
    :param dx_obj: Delphix session object from config
    :type dx_obj: lib.get_session.GetSession object
    :param engine: name of an engine, all or None
    :type engine: str
    :param single_thread: Run as single thread (True) or
                    multiple threads (False)
    :type single_thread: bool
    """
    threads = []
    # if engine ="all", run against every engine in config_file
    if engine == "all":
        dx_logging.print_info(f"Executing on all Delphix Engines")
        try:
            for delphix_ddp in dx_obj.dlpx_ddps:
                engine_ref = dx_obj.dlpx_ddps[delphix_ddp]
                dx_obj.jobs[engine_ref["ip_address"]] = []

                t = _get_thread(
                    main_func, engine_ref, dx_obj, single_thread, helper_ref
                )
                threads.append(t)
                # TODO: Revisit threading logic
                # This sleep has been tactically added to prevent errors in the parallel
                # processing of operations across multiple engines
                time.sleep(2)
        except dlpx_exceptions.DlpxException as err:
            dx_logging.print_exception(
                f"Error encountered in run_job():\n{err}"
            )
            raise err
    elif engine == "default":
        try:
            for delphix_ddp in dx_obj.dlpx_ddps.keys():
                is_default = dx_obj.dlpx_ddps[delphix_ddp]["default"]
                if is_default and is_default.lower() == "true":
                    dx_obj_default = dx_obj
                    dx_obj_default.dlpx_ddps = {
                        delphix_ddp: dx_obj.dlpx_ddps[delphix_ddp]
                    }
                    engine_ref = dx_obj.dlpx_ddps[delphix_ddp]
                    dx_obj.jobs[engine_ref["ip_address"]] = []
                    dx_logging.print_info(
                        f"Executing on the default Delphix Engine"
                    )
                    t = _get_thread(
                        main_func,
                        engine_ref,
                        dx_obj,
                        single_thread,
                        helper_ref,
                    )
                    threads.append(t)
                    break
        except TypeError as err:
            raise dlpx_exceptions.DlpxException(f"Error in run_job: {err}")
        except (dlpx_exceptions.DlpxException) as e:
            dx_logging.print_exception(f"Error in run_job():\n{e}")
            raise e
    else:
        # Test to see if the engine exists in config_file
        try:
            engine_ref = dx_obj.dlpx_ddps[engine]
            dx_obj.jobs[engine_ref["ip_address"]] = []
            t = _get_thread(
                main_func, engine_ref, dx_obj, single_thread, helper_ref
            )
            threads.append(t)
            dx_logging.print_info(
                f"Executing on Delphix Engine: " f'{engine_ref["hostname"]}'
            )
        except (exceptions.RequestError, KeyError):
            raise dlpx_exceptions.DlpxException(
                f"\nERROR: Delphix Engine: {engine} cannot be found. Please "
                f"check your input and try again."
            )
        except (dlpx_exceptions.DlpxException) as e:
            dx_logging.print_exception(f"Error in run_job():\n{e}")
            raise e
    if engine is None:
        raise dlpx_exceptions.DlpxException(
            f"ERROR: No default Delphix Engine found."
        )
    return threads


def track_running_jobs(
    engine, dx_obj, poll=10, failures=None, with_progress_bar=True
):
    """
    Retrieves running job state
    :param engine: Dictionary containing info on the DDP (IP, username, etc.)
    :param poll: How long to sleep between querying jobs
    :param dx_obj: Delphix session object from config
    :type dx_obj: lib.get_session.GetSession object
    :type poll: int
    :return:
    """
    # get all the jobs, then inspect them
    if failures is None:
        failures = [False]
    engine_running_jobs = ""
    if engine["ip_address"] in dx_obj.jobs:
        engine_running_jobs = dx_obj.jobs[engine["ip_address"]]
    if not engine_running_jobs:
        dx_logging.print_debug(
            f'No running jobs on engine : {engine["hostname"]}'
        )
    else:
        dx_logging.print_info(f'Waiting for jobs on : {engine["hostname"]}')
    while engine_running_jobs:
        for j in engine_running_jobs:
            job_obj = job.get(dx_obj.server_session, j)
            if job_obj.job_state in ["COMPLETED"]:
                # util.show_progress(100, job_obj.reference, with_progress_bar=with_progress_bar)
                engine_running_jobs.remove(j)
                dx_logging.print_info(
                    f'Engine: {engine["hostname"]}: {job_obj.reference} is 100% COMPLETE'
                )
            elif job_obj.job_state in ["CANCELED", "FAILED"]:
                err_msg = extract_failure_message(job_obj)
                engine_running_jobs.remove(j)
                dx_logging.print_exception(
                    f'Engine: {engine["hostname"]}: {job_obj.reference} was CANCELLED or FAILED '
                    f"due to an error"
                )
                dx_logging.print_exception(f"{err_msg}")
                failures[0] = True
                raise Exception(f"{job_obj.job_id} {job_obj.job_state}")

            elif job_obj.job_state in "RUNNING":
                # util.show_progress(job_obj.percent_complete, job_obj.reference, with_progress_bar=with_progress_bar)
                dx_logging.print_info(
                    f'Engine: {engine["hostname"]}: {job_obj.reference} is RUNNING and {job_obj.percent_complete}% complete '
                )
        if engine_running_jobs:
            time.sleep(poll)


def extract_failure_message(jobobj):
    try:
        events = jobobj.events
        err_event = [event for event in events if event.event_type == "ERROR"]
        if err_event:
            err_msg = f"Error Reason: {err_event[0].message_details} \n Possible Action: {err_event[0].message_action}"
        return err_msg
    except (Exception) as err:
        return


def find_job_state(engine, dx_obj, poll=5):
    """
    Retrieves running job state
    :param engine: Dictionary containing info on the DDP (IP, username, etc.)
    :param poll: How long to sleep between querying jobs
    :param dx_obj: Delphix session object from config
    :type dx_obj: lib.get_session.GetSession object
    :type poll: int
    :return:
    """
    # get all the jobs, then inspect them
    dx_logging.print_debug(f"Checking running jobs state")
    i = 0
    for j in dx_obj.jobs.keys():
        print(len(dx_obj.jobs), j)
        job_obj = job.get(dx_obj.server_session, dx_obj.jobs[j])
        dx_logging.print_debug(
            f'{engine["ip_address"]}: Running job: ' f"{job_obj.job_state}"
        )
        if job_obj.job_state in ["CANCELED", "COMPLETED", "FAILED"]:
            # If the job is in a non-running state, remove it
            # from the running jobs list.
            del dx_obj.jobs[j]
            if len(dx_obj.jobs) == 0:
                break
        elif job_obj.job_state in "RUNNING":
            # If the job is in a running state, increment the
            # running job count.
            i += 1
        dx_logging.print_debug(f'{engine["ip_address"]}: {i} jobs running.')
        # If we have running jobs, pause before repeating the
        # checks.
        if dx_obj.jobs:
            time.sleep(poll)
        else:
            dx_logging.print_debug(f"No jobs running")
            break


def find_job_state_by_jobid(engine, dx_obj, job_id, poll=20):
    """
    Retrieves running job state
    :param engine: Dictionary containing info on the DDP (IP, username, etc.)
    :param poll: How long to sleep between querying jobs
    :param dx_obj: Delphix session object from config
    :type dx_obj: lib.get_session.GetSession object
    :param job_id: Job ID to check the state
    :type poll: int
    :return:
    """
    # get the job object
    job_obj = job.get(dx_obj.server_session, job_id)
    dx_logging.print_debug(job_obj)
    dx_logging.print_debug(f" Waiting for : {job_id} to finish")
    while job_obj.job_state == "RUNNING":
        time.sleep(poll)
        job_obj = job.get(dx_obj.server_session, job_id)
    dx_logging.print_debug(
        f"Job: {job_id} completed with status: {job_obj.job_state}"
    )
    return job_obj.job_state


def time_elapsed(time_start):
    """
    This function calculates the time elapsed since the beginning of the script.
    Call this anywhere you want to note the progress in terms of time
    :param time_start: start time of the script.
    :type time_start: float
    """
    return round((time.time() - time_start) / 60, +1)



"""
Custom exception class for delphixpy scripts
"""


class DlpxException(BaseException):
    """
    Delphix Exception class. Exit signals are handled by calling method.
    """

    def __init__(self, error):
        super(DlpxException, self).__init__(error)
        self._error = error

    @property
    def error(self):
        """
        Return an DlpxException object describing this error.
        """
        return self.error


class DlpxObjectNotFound(BaseException):
    """
    Delphix Exception class. Exit signals are handled by calling method.
    Raised when a Delphix Object is not found
    """

    def __init__(self, message):
        super(DlpxObjectNotFound, self).__init__(message)
        self._message = message

    @property
    def message(self):
        """
        Return an ErrorResult object describing this request message.
        """
        return self._message


class DlpxObjectExists(BaseException):
    """
    Delphix Exception class. Exit signals are handled by calling method.
    Raised when a Delphix Object is found
    """

    def __init__(self, message):
        super(DlpxObjectExists, self).__init__(message)
        self._message = message

    @property
    def message(self):
        """
        Return an ErrorResult object describing this request message.
        """
        return self._message



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#
"""
    Delpix Integration (DXI) module for Self Service Bookmarks
    [create | update | share | unshare | list | delete]
"""

from delphixpy.v1_10_2 import exceptions
from delphixpy.v1_10_2.web import selfservice
from delphixpy.v1_10_2.web import vo
from dxi._lib import dlpx_exceptions as dxe
from dxi._lib import dx_logging as log
from dxi._lib import get_references as ref
from dxi._lib import run_job
from dxi._lib.dxi_constants import BookmarkOps
from dxi._lib.dxi_constants import DataLayoutType
from dxi._lib.run_async import run_async
from dxi._lib.util import DXIConfigConstantsHelper
from dxi.dxi_tool_base import DXIBase
from tabulate import tabulate


class BookmarkConstants(object):
    """
    Class of common Bookmark Constants
    """

    SINGLE_THREAD = False
    POLL = 20
    CONFIG = DXIConfigConstantsHelper().get_config()
    LOG_FILE_PATH = DXIConfigConstantsHelper().get_logdir()
    ENGINE_ID = "default"
    PARALLEL = 5
    TYPE = None
    ACTION = None
    HOSTIP = None
    MODULE_NAME = __name__
    LIST_HEADER = [
        "Bookmark Name",
        "Bookmark Reference",
        "Branch",
        "Bookmark Type",
        "DataLayout Name",
        "Tags",
    ]


class DXIBookmark(DXIBase):
    """
    Delphix Integration class for Bookmark Operations

    This class contains all methods to perform Delphix \
    Self Service Bookmark Operations
    """

    def __init__(
        self,
        engine=BookmarkConstants.ENGINE_ID,
        log_file_path=BookmarkConstants.LOG_FILE_PATH,
        config=BookmarkConstants.CONFIG,
        poll=BookmarkConstants.POLL,
        single_thread=BookmarkConstants.SINGLE_THREAD,
        parallel=BookmarkConstants.PARALLEL,
        action=BookmarkConstants.ACTION,
        module_name=BookmarkConstants.MODULE_NAME,
    ):
        """
        :param engine: Identifier for the Delphix engine in dxtools.conf
        :type engine: `str`

        :param log_file: Log file for the module
        :type log_file: `str`

        :param config: Location (including filename) of the config file
        :type config: `str`

        :param poll: The number of seconds to wait between job polls
        :type poll: `str`

        :param single_thread: Run as single thread.
                              False - To run as multiple threads
                              True - To run as single thead [default]
        :type single_thread: `str`

        :param parallel: Limit number of jobs to maxjob
        :type parallel: int`

        :param action: Type of the action being performed
        :type config: `str`

        :param module_name: Name of the current module
        :type poll: `str`

        """
        super().__init__(
            parallel=parallel,
            poll=poll,
            config=config,
            log_file_path=log_file_path,
            single_thread=single_thread,
            module_name=module_name,
            engine=engine,
        )
        self.action = action
        self.tags = ""
        self.bookmark_name = ""
        self.container_name = ""
        self.template_name = ""
        self.branch_name = ""
        self.description = ""
        self.timestamp = ""
        self.expires = ""
        self.new_bookmark_name = ""

    def list(self, tags=None):
        """
        List all existing Bookmarks
        """
        self.action = BookmarkOps.LIST
        self.tags = tags
        try:
            self._execute_operation(self._list_bookmark_helper)
            return True
        except Exception:
            return False

    def create(
        self,
        bookmark_name,
        container_name,
        template_name,
        branch_name=None,
        timestamp=None,
        expires=None,
        tags=None,
        description=None,
    ):
        """
        :param bookmark_name:
        :param container_name:
        :param template_name:
        :param branch_name:
        :param timestamp: Timestamp to create the bookmark from
                          Format: %Y-%m-%dT%H:%M:%S
                          Sample: 2021-02-23T00:00:00
        :param expires:
        :param tags:
        :param description:
        :return:
        """
        self.action = BookmarkOps.CREATE
        self.tags = tags
        self.bookmark_name = bookmark_name
        self.container_name = container_name
        self.template_name = template_name
        self.branch_name = branch_name
        self.description = description
        self.timestamp = timestamp
        self.expires = expires
        try:
            self._execute_operation(self._create_bookmark_helper)
            return True
        except Exception:
            return False

    def share(self, bookmark_name):
        """
        Share a bookmark by name
        """
        self.action = BookmarkOps.SHARE
        self.bookmark_name = bookmark_name
        try:
            self._execute_operation(self._share_bookmark_helper)
            return True
        except Exception:
            return False

    def unshare(self, bookmark_name):
        """
        Unshare a bookmark by name
        """
        self.action = BookmarkOps.UNSHARE
        self.bookmark_name = bookmark_name
        try:
            self._execute_operation(self._unshare_bookmark_helper)
            return True
        except Exception:
            return False

    def update(
        self,
        bookmark_name,
        tags=None,
        expires=None,
        new_bookmark_name=None,
        description=None,
    ):
        """
        Share a bookmark by name
        """
        self.action = BookmarkOps.UPDATE
        self.bookmark_name = bookmark_name
        self.tags = tags
        self.expires = expires
        self.new_bookmark_name = new_bookmark_name
        self.description = description
        try:
            self._execute_operation(self._update_bookmark_helper)
            return True
        except Exception:
            return False

    def delete(self, bookmark_name):
        """
        Delete a bookmark by name
        """
        self.action = BookmarkOps.DELETE
        self.bookmark_name = bookmark_name
        try:
            self._execute_operation(self._delete_bookmark_helper)
            return True
        except Exception:
            return False

    @run_async
    def _list_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        List all bookmarks on a specific engine

        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`

        :return: list of bookmarks on the engine
        :rtype: `list`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        tag_filter = None
        display_list = []
        tag = ""
        try:
            with dlpx_obj.job_mode(self.single_thread):
                bookmark_list = []
                ss_bookmarks = selfservice.bookmark.get_all(
                    dlpx_obj.server_session
                )
                for ss_bookmark in ss_bookmarks:
                    branch_name = ref.find_obj_name(
                        dlpx_obj.server_session,
                        selfservice.branch,
                        ss_bookmark.branch,
                    )
                    data_layout = (
                        ss_bookmark.container_name
                        if ss_bookmark.container_name is not None
                        else ss_bookmark.template_name
                    )
                    if self.tags:
                        tag_filter = [x.strip() for x in self.tags.split(",")]
                    if tag_filter is None:
                        tag = ss_bookmark.tags if ss_bookmark.tags else None
                        if tag:
                            tag = ", ".join(tag for tag in ss_bookmark.tags)
                        display_list.append(
                            [
                                ss_bookmark.name,
                                ss_bookmark.reference,
                                branch_name,
                                ss_bookmark.bookmark_type,
                                data_layout,
                                tag,
                            ]
                        )
                        bookmark_list.append(
                            dict(
                                zip(
                                    BookmarkConstants.LIST_HEADER,
                                    [
                                        ss_bookmark.name,
                                        ss_bookmark.reference,
                                        branch_name,
                                        ss_bookmark.bookmark_type,
                                        data_layout,
                                        tag,
                                    ],
                                )
                            )
                        )
                    elif all(tag in ss_bookmark.tags for tag in tag_filter):
                        display_list.append(
                            [
                                ss_bookmark.name,
                                ss_bookmark.reference,
                                branch_name,
                                ss_bookmark.bookmark_type,
                                data_layout,
                                ",".join(tag for tag in ss_bookmark.tags),
                            ]
                        )
                        bookmark_list.append(
                            dict(
                                zip(
                                    BookmarkConstants.LIST_HEADER,
                                    [
                                        ss_bookmark.name,
                                        ss_bookmark.reference,
                                        branch_name,
                                        ss_bookmark.bookmark_type,
                                        data_layout,
                                        tag,
                                    ],
                                )
                            )
                        )
                print("\n")
                print(
                    tabulate(
                        display_list,
                        headers=BookmarkConstants.LIST_HEADER,
                        tablefmt="grid",
                    )
                )
            self._add_last_job_to_track(dlpx_obj)
            run_job.track_running_jobs(
                engine_ref, dlpx_obj, poll=self.poll, failures=self.failures
            )
            return bookmark_list
        except (
            dxe.DlpxException,
            exceptions.HttpError,
            exceptions.RequestError,
        ) as err:
            log.print_exception(
                f"ERROR: The bookmarks could not be" f"listed. ERROR: {err}"
            )

    @run_async
    def _create_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        Creating a bookmark

        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`

        :return: created bookmark reference
        :rtype: `str`
        """
        bookmark_ref = None
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)

        ss_bookmark_params = vo.JSBookmarkCreateParameters()
        ss_bookmark_params.bookmark = vo.JSBookmark()
        ss_bookmark_params.bookmark.name = self.bookmark_name
        data_layout = (
            self.template_name if self.template_name else self.container_name
        )
        layout_type = (
            DataLayoutType.DATA_TEMPLATE.value
            if self.template_name
            else DataLayoutType.DATA_CONTAINER.value
        )
        engine_addr = dlpx_obj.server_session.address
        try:
            if self.template_name:
                data_layout_obj = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.template,
                    self.template_name,
                )
            elif self.container_name:
                data_layout_obj = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.container,
                    self.container_name,
                )
            else:
                log.print_exception(
                    f"A containername or templatename must "
                    f"be provided to create a bookmark"
                )
                self.failures[0] = True
                raise dxe.DlpxException(
                    f"A container or branch much "
                    f"be provided to create a bookmark"
                )
        except dxe.DlpxObjectNotFound:
            log.print_exception(
                f"Unable to find a container or template on engine"
            )
            self.failures[0] = True
            raise dxe.DlpxObjectNotFound(
                "Unable to find a container or template on engine"
            )

        if self.branch_name:
            for branch_obj in selfservice.branch.get_all(
                dlpx_obj.server_session
            ):
                if (
                    self.branch_name == branch_obj.name
                    and data_layout_obj.reference == branch_obj.data_layout
                ):
                    ss_bookmark_params.bookmark.branch = branch_obj.reference
                    break
            if ss_bookmark_params.bookmark.branch is None:
                log.print_exception(
                    f"{self.branch_name} was not found in the "
                    f"{layout_type}:{data_layout}"
                )
                raise dxe.DlpxException(
                    f"{self.branch_name} was not found in the "
                    f"{layout_type}:{data_layout}"
                )
        elif not self.branch_name:
            try:
                ss_bookmark_params.bookmark.branch = (
                    data_layout_obj.active_branch
                )
            except (
                dxe.DlpxException,
                exceptions.RequestError,
                dxe.DlpxObjectNotFound,
            ) as err:
                log.print_exception(
                    f"Could not find a default branch for "
                    f"{self.data_layout_type}:{self.data_layout_name} "
                    f"on engine: {engine_addr} "
                )
                self.failures[0] = True
                raise err

        if self.tags:
            ss_bookmark_params.bookmark.tags = self.tags.split(",")
        if self.description:
            ss_bookmark_params.bookmark.description = self.description
        if self.expires:
            ss_bookmark_params.bookmark.expiration = ref.convert_timestamp_toiso(  # noqa
                dlpx_obj.server_session, self.expires
            )

        ss_bookmark_params.timeline_point_parameters = (
            vo.JSTimelinePointLatestTimeInput()
        )
        ss_bookmark_params.timeline_point_parameters.source_data_layout = (
            data_layout_obj.reference
        )

        if self.timestamp and self.timestamp.lower() != "latest":
            ss_bookmark_params.timeline_point_parameters = (
                vo.JSTimelinePointTimeInput()
            )
            ss_bookmark_params.timeline_point_parameters.branch = (
                ss_bookmark_params.bookmark.branch
            )
            ss_bookmark_params.timeline_point_parameters.time = ref.convert_timestamp_toiso(  # noqa
                dlpx_obj.server_session, self.timestamp
            )

        try:
            with dlpx_obj.job_mode(self.single_thread):
                bookmark_ref = selfservice.bookmark.create(
                    dlpx_obj.server_session, ss_bookmark_params
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
            return bookmark_ref
        except (
            dxe.DlpxException,
            exceptions.RequestError,
            exceptions.HttpError,
        ) as err:
            log.print_exception(
                f"The bookmark {self.bookmark_name} was not "
                f"created. ERROR :{err}"
            )
            self.failures[0] = True
            # raise dxe.DlpxException(
            #     f"The bookmark {self.bookmark_name} was not "
            #     f"created. ERROR :{err}"
            # )

    @run_async
    def _unshare_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        UnShare a bookmark
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`

        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        try:
            with dlpx_obj.job_mode(self.single_thread):
                bmk_ref = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.bookmark,
                    self.bookmark_name,
                ).reference

                selfservice.bookmark.unshare(dlpx_obj.server_session, bmk_ref)
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
        except (exceptions.HttpError, exceptions.RequestError) as err:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} could not be unshared. "
                f"ERR: {err}"
            )
            self.failures[0] = True
            # raise err
        except dxe.DlpxObjectNotFound as err:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} Not found: {err}"
            )
            self.failures[0] = True
            # raise err
        except Exception as err:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} could not be unshared. "
                f"ERR: {err}"
            )
            self.failures[0] = True
            # raise err

    @run_async
    def _share_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        Share a bookmark

        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        try:
            with dlpx_obj.job_mode(self.single_thread):
                bmk_ref = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.bookmark,
                    self.bookmark_name,
                ).reference

                selfservice.bookmark.share(dlpx_obj.server_session, bmk_ref)
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
        except (exceptions.HttpError, exceptions.RequestError) as err:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} could not be shared. "
                f"ERR: {err}"
            )
            self.failures[0] = True
        except dxe.DlpxObjectNotFound:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} Not found"
            )
            self.failures[0] = True
        except Exception as err:
            log.print_exception(
                f"ERROR: Bookmark {self.bookmark_name} could not be shared. "
                f"ERR: {err}"
            )
            self.failures[0] = True

    @run_async
    def _update_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        Updates a bookmark
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        """
        ss_bookmark_obj = vo.JSBookmark()
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        try:
            with dlpx_obj.job_mode(self.single_thread):
                bmk_ref = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.bookmark,
                    self.bookmark_name,
                ).reference
                if self.tags:
                    ss_bookmark_obj.tags = self.tags.split(",")
                if self.description:
                    ss_bookmark_obj.description = self.description
                if self.new_bookmark_name:
                    ss_bookmark_obj.name = self.new_bookmark_name
                if self.expires:
                    ss_bookmark_obj.expiration = ref.convert_timestamp_toiso(  # noqa
                        dlpx_obj.server_session, self.expires
                    )
                selfservice.bookmark.update(
                    dlpx_obj.server_session, bmk_ref, ss_bookmark_obj
                )
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
        except (exceptions.HttpError, exceptions.RequestError) as err:
            log.print_exception(
                f"ERROR: {self.bookmark_name} could not be updated. "
                f"ERR: {err}"
            )
            self.failures[0] = True
        except dxe.DlpxObjectNotFound:
            log.print_exception(f"ERROR: {self.bookmark_name} Not found")
            self.failures[0] = True
        except Exception as err:
            log.print_exception(
                f"ERROR: {self.bookmark_name} could not be updated. "
                f"ERR: {err}"
            )
            self.failures[0] = True

    @run_async
    def _delete_bookmark_helper(self, engine_ref, dlpx_obj, single_thread):
        """
        Deletes a bookmark
        :param engine_ref: Dictionary of engines
        :type engine_ref: `dict`

        :param dlpx_obj: DDP session object
        :type dlpx_obj: `lib.GetSession.GetSession`
        """
        dlpx_obj = self._initialize_session()
        self._setup_dlpx_session(dlpx_obj, engine_ref)
        try:
            with dlpx_obj.job_mode(self.single_thread):
                bmk_ref = ref.find_obj_by_name(
                    dlpx_obj.server_session,
                    selfservice.bookmark,
                    self.bookmark_name,
                ).reference
                selfservice.bookmark.delete(dlpx_obj.server_session, bmk_ref)
                self._add_last_job_to_track(dlpx_obj)
                run_job.track_running_jobs(
                    engine_ref,
                    dlpx_obj,
                    poll=self.poll,
                    failures=self.failures,
                )
        except (exceptions.HttpError, exceptions.RequestError) as err:
            log.print_exception(
                f"ERROR: {self.bookmark_name} could not be deleted. "
                f"ERR: {err}"
            )
            self.failures[0] = True
        except dxe.DlpxObjectNotFound:
            log.print_exception(f"ERROR: {self.bookmark_name} Not found")
            self.failures[0] = True
        except Exception as err:
            log.print_exception(
                f"ERROR: {self.bookmark_name} could not be deleted. "
                f"ERR: {err}"
            )
            self.failures[0] = True



#
# Copyright (c) 2021 by Delphix. All rights reserved.
#

import pathlib

from setuptools import find_packages
from setuptools import setup

here = pathlib.Path(__file__).parent.resolve()
# Get the long description from the README file
long_description = (here / "pypidesc.md").read_text(encoding="utf-8")
__version__ = open("dxi/_version.py").readlines()[-1].split()[-1].strip("\"'")
setup(
    version=__version__,
    packages=find_packages(),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Database",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    install_requires=[
        "click",
        "certifi",
        "delphixpy",
        "python-dateutil",
        "colorama",
        "tabulate",
        "cryptography",
        "pytz"
    ],
    # Format is mypkg.mymodule:the_function'
    entry_points="""
        [console_scripts]
        dxi=dxi.dxi:dxi
    """,
    author="sunny",
    keywords="automation",  # noqa
    license="Apache 2",
    description="test",
    dependency_links=[],
    name="snytest",  #
    long_description_content_type="text/markdown",
    long_description=long_description,
    include_package_data=True,
    project_urls={},  # Optional
)
